# -*- coding: utf-8 -*-
# This file is part of convertdate.
# http://github.com/fitnr/convertdate
# Licensed under the MIT license:
# http://opensource.org/licenses/MIT
# Copyright (c) 2016, fitnr <fitnr@fakeisthenewreal>
"""
The `Julian day <https://en.wikipedia.org/wiki/Julian_day>`__
is a continuous count of days since the beginning of the Julian era on January 1, 4713 BC.
"""
from datetime import datetime

from pytz import utc

from . import gregorian, julian


def to_datetime(jdc):
    '''Return a datetime for the input floating point julian day count'''
    year, month, day = gregorian.from_jd(jdc)

    # in jdc: 0.0 = noon, 0.5 = midnight
    # the 0.5 changes it to 0.0 = midnight, 0.5 = noon
    frac = (jdc + 0.5) % 1

    hours = int(24 * frac)

    mfrac = frac * 24 - hours
    mins = int(60 * round(mfrac, 6))

    sfrac = mfrac * 60 - mins
    secs = int(60 * round(sfrac, 6))

    msfrac = sfrac * 60 - secs

    # down to ms, which are 1/1000 of a second
    ms = int(1000 * round(msfrac, 6))

    return datetime(year, month, day, int(hours), int(mins), int(secs), int(ms), tzinfo=utc)


def from_datetime(dt):
    '''Convert from ``datetime`` to julian day count.'''
    # take account of offset (if there isn't one, act like it's utc)
    try:
        dt = dt + dt.utcoffset()
    except TypeError:
        # Assuming UTC
        pass

    jdc = gregorian.to_jd(dt.year, dt.month, dt.day)

    hfrac = dt.hour / 24.0
    mfrac = round(dt.minute / (24.0 * 60), 5)
    sfrac = round(dt.second / (24.0 * 60 * 60), 5)
    msfrac = dt.microsecond / (24.0 * 60 * 60 * 1000)

    return jdc + hfrac + mfrac + sfrac + msfrac


def to_gregorian(jdc):
    '''Convert from julian day count to Gregorian date.'''
    return gregorian.from_jd(jdc)


def from_gregorian(year, month, day):
    '''Convert from Gregorian ``year``, ``month`` and ``day`` to julian day count.'''
    return gregorian.to_jd(year, month, day)


def to_julian(jdc):
    '''Convert from julian day count to Julian date.'''
    return julian.from_jd(jdc)


def from_julian(year, month, day):
    '''Convert from Julian ``year``, ``month`` and ``day`` to julian day count.'''
    return julian.to_jd(year, month, day)



# -*- coding: utf-8 -*-
# This file is part of convertdate.
# http://github.com/fitnr/convertdate
# Licensed under the MIT license:
# http://opensource.org/licenses/MIT
# Copyright (c) 2016, fitnr <fitnr@fakeisthenewreal>
import itertools
from math import trunc

from . import gregorian
from .utils import amod

EPOCH = 584282.5
HAAB = [
    "Pop",
    "Wo'",
    "Zip",
    "Sotz'",
    "Sek",
    "Xul",
    "Yaxk'in'",
    "Mol",
    "Ch'en",
    "Yax",
    "Sak'",
    "Keh",
    "Mak",
    "K'ank'in",
    "Muwan'",
    "Pax",
    "K'ayab",
    "Kumk'u",
    "Wayeb'",
]

HAAB_TRANSLATIONS = [
    "Mat",
    "Frog",
    "Red",
    "Bat",
    "Bee",
    "Dog",
    "First Sun",
    "Water",
    "Cave",
    "Green",
    "White",
    "Red",
    "Encloser",
    "Yellow Sun",
    "Screech Owl",
    "Planting Time",
    "Turtle",
    "Ripe Corn",
    "Nameless",
]

TZOLKIN = [
    "Imix'",
    "Ik'",
    "Ak'b'al",
    "K'an",
    "Chikchan",
    "Kimi",
    "Manik'",
    "Lamat",
    "Muluk",
    "Ok",
    "Chuwen",
    "Eb'",
    "B'en",
    "Ix",
    "Men",
    "K'ib'",
    "Kab'an",
    "Etz'nab'",
    "Kawak",
    "Ajaw",
]

TZOLKIN_TRANSLATIONS = [
    "Water",
    "Wind",
    "Darkness",
    "Net",
    "Feathered Serpent",
    "Death",
    "Deer",
    "Seed",
    "Jade",
    "Dog",
    "Thread",
    "Path",
    "Maize",
    "Tiger",
    "Bird",
    "Will",
    "Wisdom",
    "Obsidian Knife",
    "Thunder",
    "Sun",
]


def to_jd(baktun, katun, tun, uinal, kin):
    '''Determine Julian day from Mayan long count'''
    return EPOCH + (baktun * 144000) + (katun * 7200) + (tun * 360) + (uinal * 20) + kin


def from_jd(jd):
    '''Calculate Mayan long count from Julian day'''
    d = jd - EPOCH
    if d < 0:
        raise ValueError("Day out of range")
    baktun = trunc(d / 144000)
    d = d % 144000
    katun = trunc(d / 7200)
    d = d % 7200
    tun = trunc(d / 360)
    d = d % 360
    uinal = trunc(d / 20)
    kin = int((d % 20))

    return (baktun, katun, tun, uinal, kin)


def to_gregorian(baktun, katun, tun, uinal, kin):
    jd = to_jd(baktun, katun, tun, uinal, kin)
    return gregorian.from_jd(jd)


def from_gregorian(year, month, day):
    jd = gregorian.to_jd(year, month, day)
    return from_jd(jd)


def to_haab(jd):
    '''Determine Mayan Haab "month" and day from Julian day'''
    # Number of days since the start of the long count
    lcount = trunc(jd) + 0.5 - EPOCH
    # Long Count begins 348 days after the start of the cycle
    day = (lcount + 348) % 365

    count = day % 20
    month = trunc(day / 20)

    return int(count), HAAB[month]


def to_tzolkin(jd):
    '''Determine Mayan Tzolkin "month" and day from Julian day'''
    lcount = trunc(jd) + 0.5 - EPOCH
    day = amod(lcount + 4, 13)
    name = amod(lcount + 20, 20)
    return int(day), TZOLKIN[int(name) - 1]


def lc_to_haab(baktun, katun, tun, uinal, kin):
    jd = to_jd(baktun, katun, tun, uinal, kin)
    return to_haab(jd)


def lc_to_tzolkin(baktun, katun, tun, uinal, kin):
    jd = to_jd(baktun, katun, tun, uinal, kin)
    return to_tzolkin(jd)


def lc_to_haab_tzolkin(baktun, katun, tun, uinal, kin):
    jd = to_jd(baktun, katun, tun, uinal, kin)
    dates = to_tzolkin(jd) + to_haab(jd)
    return "{0} {1} {2} {3}".format(*dates)


def translate_haab(h):
    return dict(list(zip(HAAB, HAAB_TRANSLATIONS))).get(h)


def translate_tzolkin(tz):
    return dict(list(zip(TZOLKIN, TZOLKIN_TRANSLATIONS))).get(tz)


def _haab_count(day, month):
    '''Return the count of the given haab in the cycle. e.g. 0 Pop == 1, 5 Wayeb' == 365'''
    if day < 0 or day > 19:
        raise IndexError("Invalid day number")

    try:
        i = HAAB.index(month)
    except ValueError as err:
        raise ValueError("'{0}' is not a valid Haab' month".format(month)) from err

    return min(i * 20, 360) + day


def _tzolkin_from_count(count):
    number = amod(count, 13)
    name = TZOLKIN[count % 20 - 1]
    return number, name


def _tzolkin_count(day, name):
    if day < 1 or day > 13:
        raise IndexError("Invalid day number")

    days = set(x + day for x in range(0, 260, 13))

    try:
        n = 1 + TZOLKIN.index(name)
    except ValueError as err:
        raise ValueError("'{0}' is not a valid Tzolk'in day name".format(name)) from err

    names = set(y + n for y in range(0, 260, 20))
    return days.intersection(names).pop()


def tzolkin_generator(number=None, name=None):
    """For a given tzolkin name/number combination, return a generator
    that gives cycle, starting with the input"""

    # By default, it will start at the beginning
    number = number or 13
    name = name or "Ajaw"

    if number > 13:
        raise ValueError("Invalid day number")

    if name not in TZOLKIN:
        raise ValueError("Invalid day name")

    count = _tzolkin_count(number, name)

    ranged = itertools.chain(list(range(count, 260)), list(range(1, count)))

    for i in ranged:
        yield _tzolkin_from_count(i)


def longcount_generator(baktun, katun, tun, uinal, kin):
    '''Generate long counts, starting with input'''
    j = to_jd(baktun, katun, tun, uinal, kin)

    while True:
        yield from_jd(j)
        j = j + 1


def next_haab(month, jd):
    '''For a given haab month and a julian day count, find the next start of that month on or after the JDC'''
    if jd < EPOCH:
        raise IndexError("Input day is before Mayan epoch.")

    hday, hmonth = to_haab(jd)

    if hmonth == month:
        days = 1 - hday

    else:
        count1 = _haab_count(hday, hmonth)
        count2 = _haab_count(1, month)

        # Find number of days between haab of given jd and desired haab
        days = (count2 - count1) % 365

    # add in the number of days and return new jd
    return jd + days


def next_tzolkin(tzolkin, jd):
    '''For a given tzolk'in day, and a julian day count, find the next occurrance of that tzolk'in after the date'''
    if jd < EPOCH:
        raise IndexError("Input day is before Mayan epoch.")

    count1 = _tzolkin_count(*to_tzolkin(jd))
    count2 = _tzolkin_count(*tzolkin)

    add_days = (count2 - count1) % 260
    return jd + add_days


def next_tzolkin_haab(tzolkin, haab, jd):
    '''For a given haab-tzolk'in combination, and a Julian day count, find the next occurrance of the combination after the date'''
    # get H & T of input jd, and their place in the 18,980 day cycle
    haabcount = _haab_count(*to_haab(jd))
    haab_desired_count = _haab_count(*haab)

    # How many days between the input day and the desired day?
    haab_days = (haab_desired_count - haabcount) % 365

    possible_haab = set(h + haab_days for h in range(0, 18980, 365))

    tzcount = _tzolkin_count(*to_tzolkin(jd))
    tz_desired_count = _tzolkin_count(*tzolkin)
    # How many days between the input day and the desired day?
    tzolkin_days = (tz_desired_count - tzcount) % 260

    possible_tz = set(t + tzolkin_days for t in range(0, 18980, 260))

    try:
        return possible_tz.intersection(possible_haab).pop() + jd
    except KeyError:
        raise IndexError("That Haab'-Tzolk'in combination isn't possible")


def month_length(month):
    """Not the actual length of the month, but accounts for the 5 unlucky/nameless days"""
    if month == "Wayeb'":
        return 5
    else:
        return 20


def haab_monthcalendar(baktun=None, katun=None, tun=None, uinal=None, kin=None, jdc=None):
    '''For a given long count, return a calender of the current haab month, divided into tzolkin "weeks"'''
    if not jdc:
        jdc = to_jd(baktun, katun, tun, uinal, kin)

    haab_number, haab_month = to_haab(jdc)
    first_j = jdc - haab_number + 1

    tzolkin_start_number, tzolkin_start_name = to_tzolkin(first_j)

    gen_longcount = longcount_generator(*from_jd(first_j))
    gen_tzolkin = tzolkin_generator(tzolkin_start_number, tzolkin_start_name)

    # 13 day long tzolkin 'weeks'
    lpad = tzolkin_start_number - 1
    rpad = 13 - (tzolkin_start_number + 19 % 13)

    monlen = month_length(haab_month)

    days = [None] * lpad + list(range(1, monlen + 1)) + rpad * [None]

    def g(x, generate):
        if x is None:
            return None
        return next(generate)

    return [[(k, g(k, gen_tzolkin), g(k, gen_longcount)) for k in days[i : i + 13]] for i in range(0, len(days), 13)]


def haab_monthcalendar_prospective(haabmonth, jdc):
    '''Give the monthcalendar for the next occurance of given haab month after jdc'''
    return haab_monthcalendar(jdc=next_haab(haabmonth, jdc))



# -*- coding: utf-8 -*-
'''
    Convert between Gregorian/Julian Day and Comte's Positivist calendar.
    The Positivist calendar has 13 months and one or two festival days.
    Festival days are given as the fourteenth month.
    The Gregorian date 1789-01-01 is Positivist 0001-01-01.
'''
# This file is part of convertdate.
# http://github.com/fitnr/convertdate
# Licensed under the MIT license:
# http://opensource.org/licenses/MIT
# Copyright (c) 2016, fitnr <fitnr@fakeisthenewreal>
from calendar import isleap

from . import gregorian
from .data import positivist as data
from .utils import floor

# Positivist calendar has 13 28-day months and one festival day

EPOCH = 2374479.5

YEAR_EPOCH = 1789

DAYS_IN_YEAR = 365

MONTHS = (
    'Moses',
    'Homer',
    'Aristotle',
    'Archimedes',
    'Caesar',
    'Saint Paul',
    'Charlemagne',
    'Dante',
    'Gutenberg',
    'Shakespeare',
    'Descartes',
    'Frederic',
    'Bichat',
    '',
)


def legal_date(year, month, day):
    '''Checks if a given date is a legal positivist date'''
    try:
        assert year >= 1
        assert 0 < month <= 14
        assert 0 < day <= 28
        if month == 14:
            if isleap(year + YEAR_EPOCH - 1):
                assert day <= 2
            else:
                assert day == 1

    except AssertionError as err:
        raise ValueError("Invalid Positivist date: ({}, {}, {})".format(year, month, day)) from err

    return True


def to_jd(year, month, day):
    '''Convert a Positivist date to Julian day count.'''
    legal_date(year, month, day)
    gyear = year + YEAR_EPOCH - 1

    return (
        gregorian.EPOCH
        - 1
        + (365 * (gyear - 1))
        + floor((gyear - 1) / 4)
        + (-floor((gyear - 1) / 100))
        + floor((gyear - 1) / 400)
        + (month - 1) * 28
        + day
    )


def from_jd(jd):
    '''Convert a Julian day count to Positivist date.'''
    try:
        assert jd >= EPOCH
    except AssertionError as err:
        raise ValueError('Invalid Julian day') from err

    depoch = floor(jd - 0.5) + 0.5 - gregorian.EPOCH

    quadricent = floor(depoch / gregorian.INTERCALATION_CYCLE_DAYS)
    dqc = depoch % gregorian.INTERCALATION_CYCLE_DAYS

    cent = floor(dqc / gregorian.LEAP_SUPPRESSION_DAYS)
    dcent = dqc % gregorian.LEAP_SUPPRESSION_DAYS

    quad = floor(dcent / gregorian.LEAP_CYCLE_DAYS)
    dquad = dcent % gregorian.LEAP_CYCLE_DAYS

    yindex = floor(dquad / gregorian.YEAR_DAYS)
    year = (
        quadricent * gregorian.INTERCALATION_CYCLE_YEARS
        + cent * gregorian.LEAP_SUPPRESSION_YEARS
        + quad * gregorian.LEAP_CYCLE_YEARS
        + yindex
    )

    if yindex == 4:
        yearday = 365
        year = year - 1

    else:
        yearday = (
            depoch
            - quadricent * gregorian.INTERCALATION_CYCLE_DAYS
            - cent * gregorian.LEAP_SUPPRESSION_DAYS
            - quad * gregorian.LEAP_CYCLE_DAYS
            - yindex * gregorian.YEAR_DAYS
        )

    month = floor(yearday / 28)

    return (year - YEAR_EPOCH + 2, month + 1, int(yearday - (month * 28)) + 1)


def from_gregorian(year, month, day):
    return from_jd(gregorian.to_jd(year, month, day))


def to_gregorian(year, month, day):
    return gregorian.from_jd(to_jd(year, month, day))


def dayname(year, month, day):
    """
    Give the name of the month and day for a given date.

    Returns:
        tuple month_name, day_name
    """
    legal_date(year, month, day)

    yearday = (month - 1) * 28 + day

    if isleap(year + YEAR_EPOCH - 1):
        dname = data.DAY_NAMES_LEAP[yearday - 1]
    else:
        dname = data.DAY_NAMES[yearday - 1]

    return MONTHS[month - 1], dname


def weekday(day):
    """
    Gives the weekday (0=Monday) of a positivist month and day.
    Note that the festival month does not have a day.
    """
    return (day % 7) - 1


def festival(month, day):
    """
    Gives the festival day for a month and day.
    Returns None if inapplicable.
    """
    return data.FESTIVALS.get((month, day))



# -*- coding: utf-8 -*-
# This file is part of convertdate.
# http://github.com/fitnr/convertdate
# Licensed under the MIT license:
# http://opensource.org/licenses/MIT
# Copyright (c) 2016, fitnr <fitnr@fakeisthenewreal>
"""
The Julian calendar was implemented by Julius Caesar in 45 BC as a reformation of
the Roman calendar. It is designed to follow the solar year, with a standard year
of 365 days and a quadrennial leap year with an intercalary day (29 February).

For the first several centuries of its use, the Julian calendar did not have a
single year-numbering system. The Romans initially specific years with the names of political
leaders. Later on, different areas employed different era with various epochs.
Between the sixth and eighth centuries, western Europe adopted the Anno Domini convention.

This numbering system does not include a year 0. However, for dates before 1,
this module uses the astronomical convention of including a year 0 to simplify
mathematical comparisons across epochs. To present a date in the standard
convention, use the :meth:`julian.format` function.
"""
from .gregorian import from_jd as gregorian_from_jd
from .gregorian import to_jd as gregorian_to_jd
from .utils import floor, jwday, monthcalendarhelper

J0000 = 1721424.5  # Julian date of Gregorian epoch: 0000-01-01
J1970 = 2440587.5  # Julian date at Unix epoch: 1970-01-01
JMJD = 2400000.5  # Epoch of Modified Julian Date system

JULIAN_EPOCH = 1721423.5
J2000 = 2451545.0  # Julian day of J2000 epoch
JULIANCENTURY = 36525.0  # Days in Julian century

HAVE_30_DAYS = (4, 6, 9, 11)
HAVE_31_DAYS = (1, 3, 5, 7, 8, 10, 12)


def leap(year):
    return year % 4 == 0


def month_length(year, month):
    if month == 2:
        daysinmonth = 29 if leap(year) else 28
    else:
        daysinmonth = 30 if month in HAVE_30_DAYS else 31

    return daysinmonth


def legal_date(year, month, day):
    '''Check if this is a legal date in the Julian calendar'''
    daysinmonth = month_length(year, month)

    if not 0 < day <= daysinmonth:
        raise ValueError("Month {} doesn't have a day {}".format(month, day))

    return True


def from_jd(jd):
    '''Calculate Julian calendar date from Julian day'''
    jd += 0.5
    a = floor(jd)
    b = a + 1524
    c = floor((b - 122.1) / 365.25)
    d = floor(365.25 * c)
    e = floor((b - d) / 30.6001)
    if e < 14:
        month = floor(e - 1)
    else:
        month = floor(e - 13)
    if month > 2:
        year = floor(c - 4716)
    else:
        year = floor(c - 4715)
    day = b - d - floor(30.6001 * e)
    return (year, month, day)


def to_jd(year, month, day):
    '''Convert to Julian day using astronomical years (0 = 1 BC, -1 = 2 BC)'''
    legal_date(year, month, day)

    # Algorithm as given in Meeus, Astronomical Algorithms, Chapter 7, page 61
    if month <= 2:
        year -= 1
        month += 12

    return (floor((365.25 * (year + 4716))) + floor((30.6001 * (month + 1))) + day) - 1524.5


def from_gregorian(year, month, day):
    '''Convert a Gregorian date to a Julian date.'''
    return from_jd(gregorian_to_jd(year, month, day))


def to_gregorian(year, month, day):
    '''Convert a Julian date to a Gregorian date.'''
    return gregorian_from_jd(to_jd(year, month, day))


def monthcalendar(year, month):
    '''
    Returns a matrix representing a month’s calendar. Each row represents a week;
    days outside of the month are represented by zeros. Each week begins with Sunday.
    '''
    start_weekday = jwday(to_jd(year, month, 1))
    monthlen = month_length(year, month)
    return monthcalendarhelper(start_weekday, monthlen)


def format(year, month, day, format_string="%-d %B %y"):
    # pylint: disable=redefined-builtin
    epoch = ''
    if year <= 0:
        year = (year - 1) * -1
        epoch = ' BCE'
    d = date(year, month, day)
    return d.strftime(format_string) + epoch



# -*- coding: utf-8 -*-
# This file is part of convertdate.
# http://github.com/fitnr/convertdate
# Licensed under the MIT license:
# http://opensource.org/licenses/MIT
# Copyright (c) 2016, fitnr <fitnr@fakeisthenewreal>
"""
The modern Persian calendar, or the Solar Hijri calendar, was adopted in 1911.
It consists of twelve months of 30 or 31 days. The new year always falls on the
March equinox.
"""
from math import floor, trunc

from pymeeus.Sun import Sun
from pymeeus.Epoch import Epoch

from . import gregorian
from .utils import ceil, jwday, monthcalendarhelper, TROPICALYEAR

EPOCH = 1948320.5
WEEKDAYS = ("Doshanbeh", "Seshhanbeh", "Chaharshanbeh", "Panjshanbeh", "Jomeh", "Shanbeh", "Yekshanbeh")

MONTHS = [
    "Farvardin",
    "Ordibehesht",
    "Khordad",
    "Tir",
    "Mordad",
    "Shahrivar",
    "Mehr",
    "Aban",
    "Azar",
    "Dey",
    "Bahman",
    "Esfand",
]

HAS_31_DAYS = (1, 2, 3, 4, 5, 6)
HAS_30_DAYS = (7, 8, 9, 10, 11)


def leap(year):
    '''Is a given year a leap year in the Persian calendar ?'''
    return (to_jd(year + 1, 1, 1) - to_jd(year, 1, 1)) > 365


def equinox_jd(gyear):
    """Calculate Julian day during which the March equinox, reckoned from the
    Tehran meridian, occurred for a given Gregorian year."""
    mean_jd = Sun.get_equinox_solstice(gyear, target='spring')
    deltat_jd = mean_jd - Epoch.tt2ut(gyear, 3) / (24 * 60 * 60.)
    # Apparent JD in universal time
    apparent_jd = deltat_jd + (Sun.equation_of_time(deltat_jd)[0] / (24 * 60.))
    # Correct for meridian of Tehran + 52.5 degrees
    return floor(apparent_jd.jde() + (52.5 / 360))


def last_equinox_jd(jd):
    """Return the Julian date of spring equinox immediately preceeding the
    given Julian date."""
    guessyear = gregorian.from_jd(jd)[0]
    last_equinox = equinox_jd(guessyear)

    while last_equinox > jd:
        guessyear = guessyear - 1
        last_equinox = equinox_jd(guessyear)

    next_equinox = last_equinox - 1

    while not (last_equinox <= jd and jd < next_equinox):
        last_equinox = next_equinox
        guessyear = guessyear + 1
        next_equinox = equinox_jd(guessyear)

    return last_equinox


def jd_to_pyear(jd):
    """
    Determine the year in the Persian astronomical calendar in which a given
    Julian day falls.

    Returns:
        tuple - (Persian year, Julian day number containing equinox for this year)
    """
    lasteq = last_equinox_jd(jd)
    return round((lasteq - EPOCH) / TROPICALYEAR) + 1, lasteq


def to_jd(year, month, day):
    '''Determine Julian day from Persian date'''
    guess = (EPOCH - 1) + (TROPICALYEAR * ((year - 1) - 1))
    y0, equinox = year - 1, 0

    while y0 < year:
        y0, equinox = jd_to_pyear(guess)
        guess = equinox + TROPICALYEAR + 2

    if month <= 7:
        m = (month - 1) * 31
    else:
        m = ((month - 1) * 30) + 6

    return equinox + m + day + 0.5


def from_jd(jd):
    '''Calculate Persian date from Julian day'''
    jd = floor(jd) + 0.5
    equinox = last_equinox_jd(jd)
    year = round((equinox - EPOCH) / TROPICALYEAR) + 1
    yday = jd - (equinox + 0.5)

    if yday <= 186:
        month = ceil(yday / 31)
        day = yday - ((month - 1) * 31)
    else:
        month = ceil((yday - 6) / 30)
        day = yday - ((month - 1) * 30) - 6

    return int(year), int(month), int(day)


def from_gregorian(year, month, day):
    return from_jd(gregorian.to_jd(year, month, day))


def to_gregorian(year, month, day):
    return gregorian.from_jd(to_jd(year, month, day))


def month_length(year, month):
    if month in HAS_30_DAYS or (month == 12 and leap(year)):
        return 30
    if month in HAS_31_DAYS:
        return 31

    return 29


def monthcalendar(year, month):
    start_weekday = jwday(to_jd(year, month, 1))
    monthlen = month_length(year, month)
    return monthcalendarhelper(start_weekday, monthlen)


def format(year, month, day):
    """Convert a Persian date into a string with the format DD MONTH YYYY."""
    # pylint: disable=redefined-builtin
    return "{0:d} {1:} {2:d}".format(day, MONTHS[month - 1], year)



# -*- coding: utf-8 -*-

DAY_NAMES = (
    "Prometheus",
    "Hercules",
    "Orpheus",
    "Ulysses",
    "Lycurgus",
    "Romulus",
    "Numa",
    "Belus",
    "Sesostris",
    "Menu",
    "Cyrus",
    "Zoroaster",
    "The Druids",
    "Buddha",
    "Fuxi",
    "Laozi",
    "Mencius",
    "Theocrats of Tibet",
    "Theocrats of Japan",
    "Manco Cápac",
    "Confucius",
    "Abraham",
    "Samuel",
    "Solomon",
    "Isaiah",
    "St. John the Baptist",
    "Haroun-al-Raschid",
    "Muhammad",
    "Hesiod",
    "Tyrtéus",
    "Anacreon",
    "Pindar",
    "Sophocles",
    "Theocritus",
    "Aeschylus",
    "Scopas",
    "Zeuxis",
    "Ictinus",
    "Praxiteles",
    "Lysippus",
    "Apelles",
    "Phidias",
    "Aesop",
    "Plautus",
    "Terence",
    "Phaedrus",
    "Juvenal",
    "Lucian",
    "Aristophanes",
    "Ennius",
    "Lucretius",
    "Horace",
    "Tibullus",
    "Ovid",
    "Lucan",
    "Virgil",
    "Anaximander",
    "Anaximenes",
    "Heraclitus",
    "Anaxagoras",
    "Democritus",
    "Herodotus",
    "Thales",
    "Solon",
    "Xenophanes",
    "Empodocles",
    "Thucydides",
    "Archytas",
    "Apollonius of Tyana",
    "Pythagoras",
    "Aristippus",
    "Antisthenes",
    "Zeno",
    "Cicero",
    "Epictetus",
    "Tacitus",
    "Socrates",
    "Xenocrates",
    "Philo of Alexandria",
    "St. John the Evangelist",
    "St. Justin",
    "St. Clement of Alexandria",
    "Origen",
    "Plato",
    "Theophrastus",
    "Herophilus",
    "Erasistratus",
    "Celsus",
    "Galen",
    "Avicenna",
    "Hippocrates",
    "Euclid",
    "Aristéus",
    "Theodisius of Bithynia",
    "Hero",
    "Pappus",
    "Diophantus",
    "Apollonius",
    "Eudoxus",
    "Pytheas",
    "Aristarchus",
    "Eratos-thenes",
    "Ptolemy",
    "Albategnius",
    "Hipparchus",
    "Varro",
    "Columella",
    "Vitruvius",
    "Strabo",
    "Frontinus",
    "Plutarch",
    "Pliny the Elder",
    "Miltiades",
    "Leonidas",
    "Aristides",
    "Cimon",
    "Xenophon",
    "Phocion",
    "Themistocles",
    "Pericles",
    "Philip",
    "Demosthenes",
    "PtolemyLagus",
    "Philopoemen",
    "Polybius",
    "Alexander",
    "JuniusBrutus",
    "Camillus",
    "Fabricius",
    "Hannibal",
    "Paulus Aemilius",
    "Marius",
    "Scipio",
    "Augustus",
    "Vespasian",
    "Hadrian",
    "Antonius",
    "Papinian",
    "Alexander Severus",
    "Trajan",
    "St. Luke",
    "St. Cyprian",
    "St. Athanasius",
    "St. Jerome",
    "St. Ambrose",
    "St. Monica",
    "St. Augustine",
    "Constantine",
    "Theodosius",
    "St. Chrysostom",
    "St. Genevieve of Paris",
    "St. Pulcheria",
    "St. Gregory the Great",
    "Hildebrand",
    "St. Benedict",
    "St. Boniface",
    "St. Isidore of Seville",
    "Lanfranc",
    "Heloise",
    "the architects of the Middle Ages",
    "St. Bernard",
    "St. Francis Xavier",
    "St. Charles Borromeo",
    "St. Theresa",
    "St. Vincent de Paul",
    "Bourdaloue",
    "William Penn",
    "Bossuet",
    "Theodoric the Great",
    "Pelayo",
    "Otho the Great",
    "St. Henry",
    "Villers",
    "Don John of Austria",
    "Alfred",
    "Charles Martel",
    "El Cid",
    "Richard I",
    "Joan of Arc",
    "Albuquerque",
    "Bayard",
    "Godfrey",
    "St. Leo the Great",
    "Gerbert",
    "Peter the Hermit",
    "Suger",
    "Alexander III",
    "St. Francis of Assisi",
    "Innocent III",
    "St. Clotilde",
    "St. Bathilda",
    "St. Stephen of Hungary",
    "St. Elizabeth of Hungary",
    "Blanche of Castille",
    "St. Ferdinand III",
    "St. Louis",
    "the Troubadours",
    "Boccaccio",
    "Rabelais",
    "Cervantes",
    "La Fontain",
    "Defoe",
    "Aristo",
    "Leonardo da Vinci",
    "Michael Angelo",
    "Holbein",
    "Poussin",
    "Velásquez",
    "Teniers",
    "Raphael",
    "Froissart",
    "Camoens",
    "The Spanish Romancers",
    "Chateaubriand",
    "Walter Scott",
    "Manzoni",
    "Tasso",
    "Petrarca",
    "Thomas of Kempis",
    "Mademoiselle de Lafayette",
    "Fénélon",
    "Klopstock",
    "Byron",
    "Milton",
    "Marco Polo",
    "Jacques Coeur",
    "Vasco de Gama",
    "Napier",
    "Lacaille",
    "Cook",
    "Columbus",
    "Benvenuto Cellini",
    "Amontons",
    "Harrison",
    "Dollond",
    "Arkwright",
    "Conté",
    "Vaucanson",
    "Stevin",
    "Mariotte",
    "Papin",
    "Black",
    "Jouffroy",
    "Dalton",
    "Watt",
    "Bernard de Palissy",
    "Guglielmini",
    "Duhamel du Monceau",
    "Saussure",
    "Coulomb",
    "Carnot",
    "Montgolfier",
    "Lope de Vega",
    "Moreto",
    "Rojas",
    "Otway",
    "Lessing",
    "Goethe",
    "Calderon",
    "Tirso",
    "Vondel",
    "Racine",
    "Voltaire",
    "Metastasio",
    "Schiller",
    "Corneille",
    "Almarcon",
    "Mademoiselle de Motteville",
    "Mademoiselle de Sévigné",
    "Lesage",
    "Mademoiselle deStaal",
    "Fielding",
    "Moliere",
    "Pergolese",
    "Sacchini",
    "Gluck",
    "Beethoven",
    "Rossini",
    "Bellini",
    "Mozart",
    "Albertus Magnus",
    "Roger Bacon",
    "St. Bonaventura",
    "Ramus",
    "Montaigne",
    "Campanella",
    "St. Thomas Aquinas",
    "Hobbes",
    "Pascal",
    "Locke",
    "Vauvenargues",
    "Diderot",
    "Cabanis",
    "Lordbacon",
    "Grotius",
    "Fontenelle",
    "Vico",
    "Fréret",
    "Montesquieu",
    "Buffon",
    "Leibnitz",
    "Robertson",
    "Adam Smith",
    "Kant",
    "Condercet",
    "Joseph de Maistre",
    "Hegel",
    "Hume",
    "Marie de Molina",
    "Cosimo de Medici the Elder",
    "Philippe de Comines",
    "Isabella of Castille",
    "Charles V",
    "Henry Iv",
    "Louis Xi",
    "L'Hôpital",
    "Barneveldt",
    "Gustavus Adolphus",
    "De Witt",
    "Ruyter",
    "William III",
    "William the Silent",
    "Ximenes",
    "Sully",
    "Mazarin",
    "Colbert",
    "D'Aranda",
    "Turgot",
    "Richelieu",
    "Sidney",
    "Franklin",
    "Washington",
    "Jefferson",
    "Bolivar",
    "Francia",
    "Cromwell",
    "Copernicus",
    "Kepler",
    "Huyghens",
    "James Bernouilli",
    "Bradley",
    "Volta",
    "Galileo",
    "Vieta",
    "Wallis",
    "Clairaut",
    "Euler",
    "D'Alembert",
    "Lagrange",
    "Newton",
    "Bergmann",
    "Priestley",
    "Cavendish",
    "Guyton Morveau",
    "Berthollet",
    "Berzelius",
    "Lavoisier",
    "Harvey",
    "Boerhaave",
    "Linnaeus",
    "Haller",
    "Lamarck",
    "Broussais",
    "Gall",
    "the Dead",
    "Sainted Women",
)

leap_day_replacements = {
    0: "Cadmus",
    1: "Theseus",
    2: "Tiresias",
    7: "Semiramus",
    12: "Ossian",
    19: "Tamehameha",
    21: "Joseph",
    23: "David",
    26: "Abderrahman",
    29: "Sappho",
    32: "Euripides",
    33: "Longus",
    42: "Pilpay",
    44: "Menander",
    60: "Leucippus",
    67: "Philolaus",
    73: "Pliny the Younger",
    74: "Arrian",
    80: "St. Irenaeus",
    82: "Tertullian",
    89: "Averrhoes",
    94: "Ctesibius",
    98: "Aratus",
    99: "Nearchus",
    100: "Berosus",
    101: "Sosigenes",
    103: "Nasreddin",
    117: "Epaminondas",
    127: "Cincinnatus",
    128: "Regulus",
    131: "the Gracchi",
    133: "Maecenas",
    134: "Titus",
    135: "Nerva",
    136: "Marcus Aurelius",
    137: "Ulpian",
    138: "Aetius",
    140: "St. James",
    149: "St. Basil",
    151: "Marcian",
    154: "St. Anthony",
    155: "St. Austin",
    156: "St. Bruno",
    157: "St. Anselm",
    158: "Beatrice",
    159: "St. Benezet",
    161: "Ignatius Loyola",
    162: "Fredrick Borromeo",
    163: "St. Catharine of Siena",
    164: "Abbé de l'Epée",
    165: "Claude Fleury",
    166: "George Fox",
    170: "Henry the Fowler",
    172: "La Valette",
    173: "John Sobieski",
    176: "Tancred",
    177: "Saladin",
    178: "Marina",
    179: "Sir Walter Raleigh",
    182: "Leo IV",
    183: "Peter Damian",
    185: "St. Eligius",
    186: "Becket",
    187: "St. Dominic",
    190: "St. Mathilda of Tuscany",
    191: "Mathias Corvinus",
    194: "Alfonso X",
    197: "Chaucer",
    198: "Swift",
    200: "Burns",
    201: "Goldsmith",
    203: "Titian",
    204: "Paul Veronese",
    205: "Rembrandt",
    206: "Lesueuer",
    207: "Murillo",
    208: "Rubens",
    210: "Joinville",
    211: "Spenser",
    214: "James Fenimore Cooper",
    218: "Louis of Granada & Bunyan",
    219: "Mademoiselle de Staël",
    220: "St. Francis of Sales",
    221: "Gessner",
    222: "Élisa Mercœur & Shelly",
    224: "Chardin",
    225: "Gresham",
    226: "Magellan",
    227: "Briggs",
    228: "Delambre",
    229: "Tasman",
    232: "Wheatstone",
    233: "Pierre Leroy",
    234: "Graham",
    235: "Jacquard",
    238: "Torricelli",
    239: "Boyle",
    240: "Worcester",
    242: "Fulton",
    243: "Thilorier",
    246: "Riquet",
    247: "Bourgelat",
    248: "Bouguer",
    249: "Borda",
    250: "Vauban",
    252: "Montalvan",
    253: "Guillem de Castro",
    254: "Guevara",
    263: "Alfieri",
    267: "Mademoiselle Roland",
    268: "Lady Montagu",
    269: "Sterne",
    270: "Miss Edgeworth",
    271: "Richardson",
    273: "Palestrina",
    274: "Grétry",
    275: "Lully",
    276: "Handel",
    277: "Weber",
    278: "Donizeti",
    280: "John of Salisbury",
    281: "Raymond Lully",
    282: "Joachim",
    283: "Nicholas of Cusa",
    284: "Erasmus",
    285: "Sir Thomas More",
    287: "Spinoza",
    288: "Giordano Bruno",
    289: "Malebranche",
    290: "Mademoiselle de Lambert",
    291: "Duclos",
    292: "George Leroy",
    294: "Cujas",
    295: "Maupertuis",
    296: "Herder",
    297: "Wincklemann",
    298: "D'Aguesseau",
    299: "Oken",
    301: "Gibbon",
    302: "Dunoyer",
    303: "Fichte",
    304: "Ferguson",
    305: "Bonald",
    306: "Sophie Germain",
    310: "Guicciardini",
    312: "Sixtus V",
    323: "Oxenstiern",
    324: "Walpole",
    325: "Louis XIV",
    326: "Pombal",
    327: "Campomanes",
    329: "Lambert",
    330: "Hampden",
    331: "Kosciusko",
    332: "Madison",
    333: "Toussaint L'Ouverture",
    336: "Tycho Brahe",
    337: "Halley",
    338: "Varignon",
    339: "John Bernouilli",
    340: "Römer",
    341: "Sauveur",
    343: "Harriot",
    344: "Fermat",
    345: "Poinsot",
    346: "Monge",
    347: "Daniel Bernouilli",
    348: "Joseph Fourier",
    350: "Scheele",
    351: "Davy",
    353: "Geoffroy",
    355: "Ritter",
    357: "Charles Bell",
    358: "Stahl & Barthez",
    359: "Bernard de Jussieu",
    360: "Félix Vicq-d'Azyr",
    361: "Blainville",
    362: "Morgagni",
}

DAY_NAMES_LEAP = [leap_day_replacements.get(i, x) for i, x in enumerate(DAY_NAMES)]

FESTIVALS = {
    (1, 1): "the Great Being",
    (1, 7): "religion",
    (1, 14): "history",
    (1, 21): "nation",
    (1, 28): "community",
    (2, 7): "complete marriage",
    (2, 14): "chaste marriage",
    (2, 21): "unequal marriage",
    (2, 28): "subjective marriage",
    (3, 7): "natural fatherhood",
    (3, 14): "artificial fatherhood",
    (3, 21): "spiritual fatherhood",
    (3, 28): "temporal fatherhood",
    (4, 7): "natural filiation",
    (4, 14): "artificial filiation",
    (4, 21): "spiritual filiation",
    (4, 28): "temporal filiation",
    (5, 7): "natural brotherhood",
    (5, 14): "artificial brotherhood",
    (5, 21): "spiritual brotherhood",
    (5, 28): "temporal brotherhood",
    (6, 7): "complete permanent domesticity",
    (6, 14): "incomplete permanent domesticity",
    (6, 21): "complete passing domesticity",
    (6, 28): "incomplete passing domesticity",
    (7, 7): "animal gods",
    (7, 14): "fire gods",
    (7, 21): "sun gods",
    (7, 28): "war gods",
    (8, 7): "castes",
    (8, 14): "polytheistic arts",
    (8, 21): "polytheistic theory",
    (8, 28): "polytheistic society",
    (9, 7): "monotheistic theology",
    (9, 14): "Catholocism",
    (9, 21): "Islam",
    (9, 28): "metaphysics",
    (10, 7): "the mother",
    (10, 14): "the wife",
    (10, 21): "the daughter",
    (10, 28): "the sister",
    (11, 7): "artistic intellectuals",
    (11, 14): "scientific intellectuals",
    (11, 21): "secondary intellectual providence",
    (11, 28): "the elderly",
    (12, 7): "the bank",
    (12, 14): "commerce",
    (12, 21): "manufacturing",
    (12, 28): "agriculture",
    (13, 7): "inventors",
    (13, 14): "emotional labor",
    (13, 21): "meditation",
    (13, 28): "passive labor",
    (14, 1): "the Dead",
    (14, 2): "Sainted Women",
}



# -*- coding: utf-8 -*-
# This file is part of convertdate.
# http://github.com/fitnr/convertdate
# Licensed under the MIT license:
# http://opensource.org/licenses/MIT
# Copyright (c) 2016, fitnr <fitnr@fakeisthenewreal>
"""Convert dates between the Hijri calendar and the Gregorian and Julian calendars."""
from . import gregorian
from .utils import ceil, floor, jwday, monthcalendarhelper

EPOCH = 1948439.5
WEEKDAYS = ("al-'ahad", "al-'ithnayn", "ath-thalatha'", "al-'arb`a'", "al-khamis", "al-jum`a", "as-sabt")
MONTHS = [
    "al-Muḥarram",
    "Ṣafar",
    "Rabīʿ al-ʾAwwal",
    "Rabīʿ ath-Thānī,",
    "Jumādā al-ʾAwwal,",
    "Jumādā ath-Thāniyah,",
    "Rajab",
    "Shaʿbān",
    "Ramaḍān",
    "Shawwāl",
    "Zū al-Qaʿdah",
    "Zū al-Ḥijjah",
]
HAS_29_DAYS = (2, 4, 6, 8, 10)
HAS_30_DAYS = (1, 3, 5, 7, 9, 11)


def leap(year):
    '''Is a given year a leap year in the Islamic calendar'''
    return (((year * 11) + 14) % 30) < 11


def to_jd(year, month, day):
    '''Determine Julian day count from Islamic date'''
    return (day + ceil(29.5 * (month - 1)) + (year - 1) * 354 + floor((3 + (11 * year)) / 30) + EPOCH) - 1


def from_jd(jd):
    '''Calculate Islamic date from Julian day'''
    jd = floor(jd) + 0.5
    year = floor(((30 * (jd - EPOCH)) + 10646) / 10631)
    month = min(12, ceil((jd - (29 + to_jd(year, 1, 1))) / 29.5) + 1)
    day = int(jd - to_jd(year, month, 1)) + 1
    return (year, month, day)


def to_jd_gregorianyear(gregorianyear, islamic_month, islamic_day):
    # Gregorian year is either 578 or 623 years greater than Islamic year
    # we'll first try 622 if conversion to gregorian isn't the same
    # year that was passed to this method, then it must be 623.
    jan1 = gregorian.to_jd(gregorianyear, 1, 1)
    yi, mi, _ = from_jd(jan1)

    if mi > islamic_month:
        yi = yi + 1

    return to_jd(yi, islamic_month, islamic_day)


def from_gregorian(year, month, day):
    return from_jd(gregorian.to_jd(year, month, day))


def to_gregorian(year, month, day):
    return gregorian.from_jd(to_jd(year, month, day))


def month_length(year, month):
    if month in HAS_30_DAYS or (month == 12 and leap(year)):
        return 30

    return 29


def monthcalendar(year, month):
    start_weekday = jwday(to_jd(year, month, 1))
    monthlen = month_length(year, month)
    return monthcalendarhelper(start_weekday, monthlen)


def format(year, month, day):
    """Convert an Islamic date into a string with the format DD MONTH YYYY."""
    # pylint: disable=redefined-builtin
    return "{0:d} {1:} {2:d}".format(day, MONTHS[month - 1], year)



# -*- coding: utf-8 -*-
# This file is part of convertdate.
# http://github.com/fitnr/convertdate
# Licensed under the MIT license:
# http://opensource.org/licenses/MIT
# Copyright (c) 2016, fitnr <fitnr@fakeisthenewreal>
"""
The Gregorian calendar was introduced by Pope Gregory XII in October 1582. It reforms
the Julian calendar by adjusting leap year rules to reduce the drift versus solar
year.

The Gregorian calendar, like the Julian, does not include a year 0. However, for dates before 1,
this module uses the astronomical convention of including a year 0 to simplify
mathematical comparisons across epochs. To present a date in the standard convention,
use the :meth:`gregorian.format` function.
"""
from calendar import isleap, monthrange
from datetime import date

from .utils import floor, jwday, monthcalendarhelper

EPOCH = 1721425.5

INTERCALATION_CYCLE_YEARS = 400
INTERCALATION_CYCLE_DAYS = 146097

LEAP_SUPPRESSION_YEARS = 100
LEAP_SUPPRESSION_DAYS = 36524

LEAP_CYCLE_YEARS = 4
LEAP_CYCLE_DAYS = 1461

YEAR_DAYS = 365

HAVE_30_DAYS = (4, 6, 9, 11)
HAVE_31_DAYS = (1, 3, 5, 7, 8, 10, 12)


def legal_date(year, month, day):
    '''Check if this is a legal date in the Gregorian calendar'''
    if month == 2:
        daysinmonth = 29 if isleap(year) else 28
    else:
        daysinmonth = 30 if month in HAVE_30_DAYS else 31

    if not 0 < day <= daysinmonth:
        raise ValueError("Month {} doesn't have a day {}".format(month, day))

    return True


def to_jd2(year, month, day):
    '''Gregorian to Julian Day Count for years between 1801-2099'''
    # http://quasar.as.utexas.edu/BillInfo/JulianDatesG.html
    legal_date(year, month, day)

    if month <= 2:
        year = year - 1
        month = month + 12

    a = floor(year / 100)
    b = floor(a / 4)
    c = 2 - a + b
    e = floor(365.25 * (year + 4716))
    f = floor(30.6001 * (month + 1))
    return c + day + e + f - 1524.5


def to_jd(year, month, day):
    '''Convert gregorian date to julian day count.'''
    legal_date(year, month, day)

    if month <= 2:
        leap_adj = 0
    elif isleap(year):
        leap_adj = -1
    else:
        leap_adj = -2

    return (
        EPOCH
        - 1
        + (YEAR_DAYS * (year - 1))
        + floor((year - 1) / LEAP_CYCLE_YEARS)
        + (-floor((year - 1) / LEAP_SUPPRESSION_YEARS))
        + floor((year - 1) / INTERCALATION_CYCLE_YEARS)
        + floor((((367 * month) - 362) / 12) + leap_adj + day)
    )


def from_jd(jd):
    '''Return Gregorian date in a (Y, M, D) tuple'''
    wjd = floor(jd - 0.5) + 0.5
    depoch = wjd - EPOCH

    quadricent = floor(depoch / INTERCALATION_CYCLE_DAYS)
    dqc = depoch % INTERCALATION_CYCLE_DAYS

    cent = floor(dqc / LEAP_SUPPRESSION_DAYS)
    dcent = dqc % LEAP_SUPPRESSION_DAYS

    quad = floor(dcent / LEAP_CYCLE_DAYS)
    dquad = dcent % LEAP_CYCLE_DAYS

    yindex = floor(dquad / YEAR_DAYS)
    year = quadricent * INTERCALATION_CYCLE_YEARS + cent * LEAP_SUPPRESSION_YEARS + quad * LEAP_CYCLE_YEARS + yindex

    if not (cent == 4 or yindex == 4):
        year += 1

    yearday = wjd - to_jd(year, 1, 1)

    leap = isleap(year)

    if yearday < 58 + leap:
        leap_adj = 0
    elif leap:
        leap_adj = 1
    else:
        leap_adj = 2

    month = floor((((yearday + leap_adj) * 12) + 373) / 367)
    day = int(wjd - to_jd(year, month, 1)) + 1

    return (year, month, day)


def month_length(year, month):
    '''Calculate the length of a month in the Gregorian calendar'''
    return monthrange(year, month)[1]


def monthcalendar(year, month):
    '''
    Return a list of lists that describe the calender for one month. Each inner
    list have 7 items, one for each weekday, starting with Sunday. These items
    are either ``None`` or an integer, counting from 1 to the number of days in
    the month.
    For Gregorian, this is very similiar to the built-in :meth:``calendar.monthcalendar``.
    '''
    start_weekday = jwday(to_jd(year, month, 1))
    monthlen = month_length(year, month)

    return monthcalendarhelper(start_weekday, monthlen)


def format(year, month, day, format_string="%-d %B %y"):
    # pylint: disable=redefined-builtin
    epoch = ''
    if year <= 0:
        year = (year - 1) * -1
        epoch = ' BCE'
    d = date(year, month, day)
    return d.strftime(format_string) + epoch



# -*- coding: utf-8 -*-
# This file is part of convertdate.
# http://github.com/fitnr/convertdate
# Licensed under the MIT license:
# http://opensource.org/licenses/MIT
# Copyright (c) 2016, fitnr <fitnr@fakeisthenewreal>
from math import floor

from . import gregorian
from .utils import jwday, monthcalendarhelper

EPOCH = 347995.5
HEBREW_YEAR_OFFSET = 3760

# Hebrew months
NISAN = 1
IYYAR = 2
SIVAN = 3
TAMMUZ = 4
AV = 5
ELUL = 6
TISHRI = 7
HESHVAN = 8
KISLEV = 9
TEVETH = 10
SHEVAT = 11
ADAR = 12
VEADAR = 13

MONTHS = [
    'Nisan',
    'Iyyar',
    'Sivan',
    'Tammuz',
    'Av',
    'Elul',
    'Tishri',
    'Heshvan',
    'Kislev',
    'Teveth',
    'Shevat',
    'Adar',
    'Adar Bet',
]

MONTHS_HEB = [
    'ניסן',
    'אייר',
    'סיוון',
    'תמוז',
    'אב',
    'אלול',
    'תשרי',
    'חשוון',
    'כסלו',
    'טבת',
    'שבט',
    'אדר',
    'אדר ב'
]


def leap(year):
    # Is a given Hebrew year a leap year ?
    return (((year * 7) + 1) % 19) < 7


def year_months(year):
    '''How many months are there in a Hebrew year (12 = normal, 13 = leap)'''
    if leap(year):
        return VEADAR

    return ADAR


def delay_1(year):
    '''Test for delay of start of the ecclesiastical new year to
    avoid improper weekdays for holidays.'''
    # Sunday, Wednesday, and Friday as start of the new year.
    months = floor(((235 * year) - 234) / 19)
    parts = 12084 + (13753 * months)
    day = (months * 29) + floor(parts / 25920)

    if ((3 * (day + 1)) % 7) < 3:
        day += 1

    return day


def delay_2(year):
    '''Check for delay in start of the ecclesiastical new year due to length
    of adjacent years'''
    last = delay_1(year - 1)
    present = delay_1(year)
    next_ = delay_1(year + 1)

    if next_ - present == 356:
        return 2

    if present - last == 382:
        return 1

    return 0


def year_days(year):
    '''How many days are in a Hebrew year ?'''
    return to_jd(year + 1, 7, 1) - to_jd(year, 7, 1)


def month_days(year, month):
    '''How many days are in a given month of a given year'''
    if month > VEADAR:
        raise ValueError("Incorrect month index")

    # First of all, dispose of fixed-length 29 day months
    if month in (IYYAR, TAMMUZ, ELUL, TEVETH, VEADAR):
        return 29

    # If it's not a leap year, Adar has 29 days
    if month == ADAR and not leap(year):
        return 29

    # If it's Heshvan, days depend on length of year
    if month == HESHVAN and (year_days(year) % 10) != 5:
        return 29

    # Similarly, Kislev varies with the length of year
    if month == KISLEV and (year_days(year) % 10) == 3:
        return 29

    # Nope, it's a 30 day month
    return 30


def to_jd(year, month, day):
    months = year_months(year)
    jd = EPOCH + delay_1(year) + delay_2(year) + day + 1

    if month < TISHRI:
        for m in range(TISHRI, months + 1):
            jd += month_days(year, m)

        for m in range(NISAN, month):
            jd += month_days(year, m)
    else:
        for m in range(TISHRI, month):
            jd += month_days(year, m)

    return int(jd) + 0.5


def from_jd(jd):
    jd = floor(jd) + 0.5
    count = floor(((jd - EPOCH) * 98496.0) / 35975351.0)
    year = count - 1
    i = count
    while jd >= to_jd(i, TISHRI, 1):
        i += 1
        year += 1

    if jd < to_jd(year, NISAN, 1):
        first = 7
    else:
        first = 1

    month = i = first
    while jd > to_jd(year, i, month_days(year, i)):
        i += 1
        month += 1

    day = int(jd - to_jd(year, month, 1)) + 1
    return (year, month, day)


def to_civil(year, month, day):
    """Convert a date in the ecclestical calendar (year starts in Nisan) to
    the civil calendar (year starts in Tishrei)."""
    if month >= TISHRI:
        year = year + 1
    return year, month, day


def to_jd_gregorianyear(gregorianyear, hebrew_month, hebrew_day):
    '''Returns the Gregorian date when a given Hebrew month and year within a given Gregorian year.'''
    # gregorian year is either 3760 or 3761 years less than hebrew year
    # we'll first try 3760 if conversion to gregorian isn't the same
    # year that was passed to this method, then it must be 3761.
    for y in (gregorianyear + HEBREW_YEAR_OFFSET, gregorianyear + HEBREW_YEAR_OFFSET + 1):
        jd = to_jd(y, hebrew_month, hebrew_day)
        gd = gregorian.from_jd(jd)
        if gd[0] == gregorianyear:
            break

        jd = None

    if not jd:  # should never occur, but just incase...
        raise ValueError("Could not determine gregorian year")

    return gregorian.to_jd(gd[0], gd[1], gd[2])


def from_gregorian(year, month, day):
    return from_jd(gregorian.to_jd(year, month, day))


def to_gregorian(year, month, day):
    return gregorian.from_jd(to_jd(year, month, day))


def monthcalendar(year, month):
    start_weekday = jwday(to_jd(year, month, 1))
    monthlen = month_days(year, month)
    return monthcalendarhelper(start_weekday, monthlen)


def format(year, month, day, lang=None):
    """Convert a Hebrew date into a string with the format DD MONTH YYYY."""
    # pylint: disable=redefined-builtin
    lang = lang or "en"
    if lang[0:2] == "he":
        month_name = MONTHS_HEB[month - 1]
    else:
        month_name = MONTHS[month - 1]
    return "{0} {1} {2}".format(day, month_name, year)



# -*- coding: utf-8 -*-
# This file is part of convertdate.
# http://github.com/fitnr/convertdate
# Licensed under the MIT license:
# http://opensource.org/licenses/MIT
# Copyright (c) 2016, fitnr <fitnr@fakeisthenewreal>
"""
Generate the dates of common North American and Jewish holidays.

Jewish holidays begin the sunset before the first (secular) day of the holiday
With the ``eve`` option set to ``True``, the day of this sunset is returned
without the ``eve`` option, the (secular) day is returned. This is the default.
"""
import calendar
import time
from math import trunc
from .utils import jwday, nth_day_of_month
from . import gregorian, hebrew, islamic, julian

# weekdays
MON = 0
TUE = 1
WED = 2
THU = 3
FRI = 4
SAT = 5
SUN = 6

# months
JAN = 1
FEB = 2
MAR = 3
APR = 4
MAY = 5
JUN = 6
JUL = 7
AUG = 8
SEP = 9
OCT = 10
NOV = 11
DEC = 12


def new_years(year, observed=None):
    '''Jan 1st, possibly observed on last day of previous year'''
    if observed:
        weekday = calendar.weekday(year, JAN, 1)
        if weekday == SAT:
            return (year - 1, DEC, 31)
        if weekday == SUN:
            return (year, JAN, 2)
    return (year, JAN, 1)


def martin_luther_king_day(year):
    '''third monday in January'''
    return nth_day_of_month(3, MON, JAN, year)


def lincolns_birthday(year):
    '''Feb 12'''
    return (year, FEB, 12)


def valentines_day(year):
    '''feb 14th'''
    return (year, FEB, 14)


def washingtons_birthday(year, observed=None):
    '''Feb 22, possibly observed on 3rd Monday in February'''
    if observed:
        return nth_day_of_month(3, MON, FEB, year)
    return (year, FEB, 22)


def presidents_day(year):
    '''3rd Monday of Feb'''
    return nth_day_of_month(3, MON, FEB, year)


def pulaski_day(year):
    '''1st monday in March'''
    return nth_day_of_month(1, MON, MAR, year)


def easter(year, church=None):
    '''Calculate Easter in the given church according to the given calendar.'''
    church = church or "western"

    if church == "western":
        return _easter_western(year)

    if church == "orthodox":
        return julian.to_gregorian(*_easter_julian(year))

    if church == "eastern":
        return julian.to_gregorian(*_easter_julian(year, mode="eastern"))

    raise ValueError("Unknown value for 'church'")


def _easter_western(year):
    '''Calculate western easter'''
    # formula taken from http://aa.usno.navy.mil/faq/docs/easter.html
    c = trunc(year / 100)
    n = year - 19 * trunc(year / 19)

    k = trunc((c - 17) / 25)

    i = c - trunc(c / 4) - trunc((c - k) / 3) + (19 * n) + 15
    i = i - 30 * trunc(i / 30)
    i = i - trunc(i / 28) * (1 - trunc(i / 28) * trunc(29 / (i + 1)) * trunc((21 - n) / 11))

    j = year + trunc(year / 4) + i + 2 - c + trunc(c / 4)
    j = j - 7 * trunc(j / 7)

    L = i - j

    month = 3 + trunc((L + 40) / 44)
    day = L + 28 - 31 * trunc(month / 4)

    date = (year, int(month), int(day))

    return date


def _easter_julian(year, mode="dionysian"):
    '''Calculate Easter for the orthodox and eastern churches in the Julian calendar.'''
    # Uses Meeus's Julian algorithm.
    meton = year % 19
    b = year % 4
    c = year % 7
    d = (19 * meton + 15) % 30
    if mode == "eastern" and meton == 0:
        d = d + 1
    e = (2 * b + 4 * c - d + 6) % 7
    fmj = 113 + d  # Easter full moon (days after -92 March)
    dmj = fmj + e + 1  # Easter Sunday (days after -92 March)
    esmj = trunc(dmj / 31)  # month of Easter Sunday
    esdj = (dmj % 31) + 1  # day of Easter Sunday

    return year, esmj, esdj


def may_day(year):
    return (year, MAY, 1)


def mothers_day(year):
    '''2nd Sunday in May'''
    return nth_day_of_month(2, SUN, MAY, year)


def memorial_day(year):
    '''last Monday in May'''
    return nth_day_of_month(0, MON, MAY, year)


def fathers_day(year):
    '''3rd Sunday in June'''
    return nth_day_of_month(3, SUN, JUN, year)


def juneteenth(year):
    '''19th of June'''
    return year, JUN, 19


def flag_day(year):
    '''June 14th'''
    return (year, JUN, 14)


def independence_day(year, observed=None):
    """Independence Day in the United States, celebrated on July 4th.
    May be observed on the previous or following day if it occurs on a Saturday
    or Sunday.

    Arguments:
        year (int): Gregorian year
        observed (boolean): If ``True``, return the date of observation.
    """
    day = 4

    if observed:
        weekday = calendar.weekday(year, JUL, 4)
        if weekday == SAT:
            day = 3
        if weekday == SUN:
            day = 5

    return (year, JUL, day)


def labor_day(year):
    '''first Monday in Sep'''
    return nth_day_of_month(1, MON, SEP, year)


def indigenous_peoples_day(year, country='usa'):
    """Celebrated on the second Monday in October in the United States."""
    if country == 'usa':
        return nth_day_of_month(2, MON, OCT, year)

    return (year, OCT, 12)


def columbus_day(year, country='usa'):
    raise DeprecationWarning("columbus_day will be removed in a future release, use indigenous_peoples_day instead")
    return indigenous_peoples_day(year, country)


def halloween(year):
    '''Halloween is celebrated on October 31st.'''
    return (year, OCT, 31)


def election_day(year):
    """In most jurisdictions in the United States, Election day occurs on
    the first Tuesday in November."""
    return nth_day_of_month(1, TUE, NOV, year)


def veterans_day(year, observed=None):
    '''Nov 11, or the following closest weekday'''
    day = 11
    if observed:
        weekday = calendar.weekday(year, NOV, 11)
        if weekday == SAT:
            day = 10
        if weekday == SUN:
            day = 12

    return (year, NOV, day)


def rememberance_day(year):
    return veterans_day(year)


def armistice_day(year):
    return veterans_day(year)


def thanksgiving(year, country='usa'):
    """In the United States, Thanksgiving is celebrated on the last Thursday
    of November. In Canada, on the second Monday of October.

    Arguments:
        year (int): Gregorian year
        country (str): either ``'usa'`` (default) or ``'canada'``
    """
    if country == 'usa':
        if year in [1940, 1941]:
            return nth_day_of_month(3, THU, NOV, year)
        elif year == 1939:
            return nth_day_of_month(4, THU, NOV, year)

        return nth_day_of_month(0, THU, NOV, year)

    if country == 'canada':
        return nth_day_of_month(2, MON, OCT, year)

    raise NotImplementedError('Unsupported argument for country')


def christmas_eve(year):
    '''The day before Christmas, or 24th of December.'''
    return (year, DEC, 24)


def christmas(year, observed=None):
    """Christmas is celebrated on the 25th of December. For the purposes of
    business closings, it may be observed on the previous or following day if
    the 25th falls on a Saturday or Sunday, respectively.

    Arguments:
        year (int): Gregorian year
        observed (boolean): If ``True``, return the date of observation.
    """
    day = 25
    if observed:
        weekday = calendar.weekday(year, DEC, 25)
        if weekday == SAT:
            day = 24
        if weekday == SUN:
            day = 26
    return (year, DEC, day)


def new_years_eve(year):
    """The last day of the Gregorian year, December. 31st."""
    return (year, DEC, 31)


def hanukkah(year, eve=None):
    jd = hebrew.to_jd_gregorianyear(year, hebrew.KISLEV, 25)
    if eve:
        jd = jd - 1
    return gregorian.from_jd(jd)


def purim(year, eve=None):
    if not hebrew.leap(year + hebrew.HEBREW_YEAR_OFFSET):
        jd = hebrew.to_jd_gregorianyear(year, hebrew.ADAR, 14)
    else:
        jd = hebrew.to_jd_gregorianyear(year, hebrew.VEADAR, 14)
    if eve:
        jd = jd - 1
    return gregorian.from_jd(jd)


def rosh_hashanah(year, eve=None):
    jd = hebrew.to_jd_gregorianyear(year, hebrew.TISHRI, 1)
    if eve:
        jd = jd - 1
    return gregorian.from_jd(jd)


def yom_kippur(year, eve=None):
    jd = hebrew.to_jd_gregorianyear(year, hebrew.TISHRI, 10)
    if eve:
        jd = jd - 1
    return gregorian.from_jd(jd)


def passover(year, eve=None):
    jd = hebrew.to_jd_gregorianyear(year, hebrew.NISAN, 15)
    if eve:
        jd = jd - 1
    return gregorian.from_jd(jd)


def shavuot(year, eve=None):
    jd = hebrew.to_jd_gregorianyear(year, hebrew.SIVAN, 6)
    if eve:
        jd = jd - 1
    return gregorian.from_jd(jd)


def sukkot(year, eve=None):
    """Sukkot, the Feast of Tabernacles or Festival of Shelters, is celebrated on
    the 15th of Tishri.
    """
    jd = hebrew.to_jd_gregorianyear(year, hebrew.TISHRI, 15)
    if eve:
        jd = jd - 1
    return gregorian.from_jd(jd)


def shemini_azeret(year, eve=None):
    jd = hebrew.to_jd_gregorianyear(year, hebrew.TISHRI, 22)
    if eve:
        jd = jd - 1
    return gregorian.from_jd(jd)


def lag_baomer(year, eve=None):
    jd = hebrew.to_jd_gregorianyear(year, hebrew.IYYAR, 18)
    if eve:
        jd = jd - 1
    return gregorian.from_jd(jd)


def tu_beshvat(year, eve=None):
    """Tu BeShvat, the 'New Year of Trees', is celebrated on the 15th of Shevat."""
    jd = hebrew.to_jd_gregorianyear(year, hebrew.SHEVAT, 15)
    if eve:
        jd = jd - 1
    return gregorian.from_jd(jd)


def tisha_bav(year, eve=None):
    """Tisha B'Av is a fast day generally celebrated on the 9th of Av, but
    sometimes postponed to the following day.
    """
    jd = hebrew.to_jd_gregorianyear(year, hebrew.AV, 9)
    if jwday(jd) == SAT:
        jd = jd + 1
    if eve:
        jd = jd - 1
    return gregorian.from_jd(jd)


# Mexican holidays


def dia_constitucion(year, observed=True):
    """
    Constitution Day, a public holiday in Mexico observed
    on the first Monday in February.
    """
    if observed:
        return nth_day_of_month(1, MON, FEB, year)

    return (year, FEB, 5)


def natalicio_benito_juarez(year, observed=True):
    """
    Benito Juárez's Birthday, a public holiday in Mexico observed
    on the third Monday in March.
    """
    if observed:
        return nth_day_of_month(3, MON, MAR, year)

    return (year, MAR, 21)


def dia_independencia(year):
    """
    Mexican independence day, observed on September 16.
    """
    return year, SEP, 16


def dia_revolucion(year):
    """
    Revolution Day, a public holiday in Mexico observed
    on the third Monday in November.
    """
    return nth_day_of_month(3, MON, NOV, year)


# Islamic holidays


def ramadan(year, eve=None):
    """The first day of Ramadan, the month of fasting in the Islamic calendar."""
    jd = islamic.to_jd_gregorianyear(year, 9, 1)
    if eve:
        jd = jd = 1
    return gregorian.from_jd(jd)


def ashura(year, eve=None):
    """Ashura is celebrated on the tenth day of Muharram, the first month in the
    Islamic calendar."""
    jd = islamic.to_jd_gregorianyear(year, 1, 10)
    if eve:
        jd = jd = 1
    return gregorian.from_jd(jd)


def eid_alfitr(year, eve=None):
    """Eid al-Fitr, the 'Festival of Breaking the Fast' is celebrated the first
    day of the month of Shawwāl."""
    jd = islamic.to_jd_gregorianyear(year, 10, 1)
    if eve:
        jd = jd = 1
    return gregorian.from_jd(jd)


def eid_aladha(year, eve=None):
    """Eid al-Adha, the 'Festival of the Sacrifice' begins on the tenth
    day of the month of Zū al-Ḥijjah."""
    jd = islamic.to_jd_gregorianyear(year, 12, 10)
    if eve:
        jd = jd = 1
    return gregorian.from_jd(jd)


class Holidays:
    '''Convenience class for fetching many holidays in a given year.'''

    # pylint: disable=missing-function-docstring

    def __init__(self, year=None):
        self.year = year or time.localtime().tm_year

    def set_year(self, year):
        self.year = year

    def __repr__(self):
        return 'Holidays({})'.format(self.year)

    # the holidays...
    @property
    def christmas(self):
        return christmas(self.year, True)

    @property
    def christmas_eve(self):
        return christmas_eve(self.year)

    @property
    def thanksgiving(self):
        return thanksgiving(self.year)

    @property
    def new_years(self):
        return new_years(self.year, True)

    @property
    def new_years_eve(self):
        return new_years_eve(self.year)

    @property
    def independence_day(self):
        return independence_day(self.year, observed=True)

    @property
    def flag_day(self):
        return flag_day(self.year)

    @property
    def election_day(self):
        return election_day(self.year)

    @property
    def presidents_day(self):
        return presidents_day(self.year)

    @property
    def washingtons_birthday(self):
        return washingtons_birthday(self.year)

    @property
    def lincolns_birthday(self):
        return lincolns_birthday(self.year)

    @property
    def memorial_day(self):
        return memorial_day(self.year)

    @property
    def juneteenth(self):
        return juneteenth(self.year)

    @property
    def labor_day(self):
        return labor_day(self.year)

    @property
    def indigenous_peoples_day(self):
        return indigenous_peoples_day(self.year)

    @property
    def columbus_day(self):
        return indigenous_peoples_day(self.year)

    @property
    def veterans_day(self):
        return veterans_day(self.year, True)

    @property
    def valentines_day(self):
        return valentines_day(self.year)

    @property
    def halloween(self):
        return halloween(self.year)

    @property
    def mothers_day(self):
        return mothers_day(self.year)

    @property
    def fathers_day(self):
        return fathers_day(self.year)

    @property
    def pulaski_day(self):
        return pulaski_day(self.year)

    @property
    def easter(self):
        return easter(self.year)

    @property
    def martin_luther_king_day(self):
        return martin_luther_king_day(self.year)

    @property
    def hanukkah(self):
        return hanukkah(self.year, eve=False)

    @property
    def purim(self):
        return purim(self.year, eve=False)

    @property
    def rosh_hashanah(self):
        return rosh_hashanah(self.year, eve=False)

    @property
    def yom_kippur(self):
        return yom_kippur(self.year, eve=False)

    @property
    def passover(self):
        return passover(self.year, eve=False)

    @property
    def shavuot(self):
        return shavuot(self.year, eve=False)

    @property
    def sukkot(self):
        return sukkot(self.year, eve=False)

    @property
    def tu_beshvat(self):
        return tu_beshvat(self.year, eve=False)

    @property
    def shemini_azeret(self):
        return shemini_azeret(self.year, eve=False)

    @property
    def lag_baomer(self):
        return lag_baomer(self.year, eve=False)

    @property
    def tisha_bav(self):
        return tisha_bav(self.year, eve=False)

    @property
    def dia_constitucion(self):
        return dia_constitucion(self.year, observed=True)

    @property
    def natalicio_benito_juarez(self):
        return natalicio_benito_juarez(self.year, observed=True)

    @property
    def dia_independencia(self):
        return dia_independencia(self.year)

    @property
    def dia_revolucion(self):
        return dia_revolucion(self.year)

    @property
    def ramadan(self):
        return ramadan(self.year)

    @property
    def ashura(self):
        return ashura(self.year)

    @property
    def eid_alfitr(self):
        return eid_alfitr(self.year)

    @property
    def eid_aladha(self):
        return eid_aladha(self.year)


if __name__ == '__main__':
    holiday = Holidays(time.localtime().tm_year)



# -*- coding: utf-8 -*-
# This file is part of convertdate.
# http://github.com/fitnr/convertdate
# Licensed under the MIT license:
# http://opensource.org/licenses/MIT
# Copyright (c) 2016, fitnr <fitnr@fakeisthenewreal>

'''
The Dublin day count is a truncated `Julian day <julian.html>`__ with an epoch of noon on December 31, 1899.

This is a convenience module, it uses :doc:`daycount` class for all functions.
'''

from . import daycount

EPOCH = 2415020  # Julian Day Count for Dublin Count 0

_dublin = daycount.DayCount(EPOCH)

to_gregorian = _dublin.to_gregorian

from_gregorian = _dublin.from_gregorian

to_jd = _dublin.to_jd

from_jd = _dublin.from_jd

from_julian = _dublin.from_julian

to_julian = _dublin.to_julian

to_datetime = _dublin.to_datetime

from_datetime = _dublin.from_datetime



# -*- coding: utf-8 -*-
# This file is part of convertdate.
# http://github.com/fitnr/convertdate
# Licensed under the MIT license:
# http://opensource.org/licenses/MIT
# Copyright (c) 2016, fitnr <fitnr@fakeisthenewreal>

"""
The Armenian calendar begins on 11 July 552 (Julian) and has two modes of
reckoning. The first is the invariant-length version consisting of 12 months
of 30 days each and five epagomenal days; the second is the version
established by Yovhannes Sarkawag in 1084, which fixed the first day of the
year with respect to the Julian calendar and added a sixth epagomenal day
every four years.

By default the invariant calendar is used, but the Sarkawag calendar can be
used beginning with the Armenian year 533 (11 August 1084) by passing the
parameter ``method='sarkawag'`` to the relevant functions.
"""

from math import trunc

from . import gregorian, julian
from .utils import jwday, monthcalendarhelper

EPOCH = 1922501.5  # beginning of proleptic year 0, day 0 of the moveable calendar
EPOCH_SARKAWAG = 2117210.5  # last day of Sarkawag's first cycle
MONTHS = [
    "nawasard",
    "hoṙi",
    "sahmi",
    "trē",
    "kʿałocʿ",
    "aracʿ",
    "mehekan",
    "areg",
    "ahekan",
    "mareri",
    "margacʿ",
    "hroticʿ",
    "aweleacʿ",
]
MONTHS_ARM = [
    "նաւասարդ",
    "հոռի",
    "սահմի",
    "տրէ",
    "քաղոց",
    "արաց",
    "մեհեկան",
    "արեգ",
    "ահեկան",
    "մարերի",
    "մարգաց",
    "հրոտից",
    "աւելեաց",
]


def _valid_date(year, month, day, method=None):
    try:
        assert 1 <= month <= 13, "Month out of range"
    except AssertionError as error:
        raise ValueError from error

    day_max = 30
    method = method or "moveable"

    if method.lower() == "sarkawag":
        year_min = 533
        if month == 13:
            day_max = 6 if leap(year) else 5

    else:
        year_min = 1
        if month == 13:
            day_max = 5

    try:
        assert year >= year_min, "Year out of range for Armenian calendar ({} method)".format(method)
    except AssertionError as err:
        raise ValueError from err

    try:
        assert 1 <= day <= day_max, "Day out of range"
    except AssertionError as err:
        raise ValueError from err

    return True


def leap(year):
    """Return true if the year was a leap year under the system of Sarkawag"""
    if year < 533:
        return False
    return year % 4 == 0


def to_jd(year, month, day, method=None):
    """Convert Armenian date to Julian day count. Use the method of Sarkawag if requested."""
    _valid_date(year, month, day, method)
    yeardays = (month - 1) * 30 + day
    if method == "sarkawag":
        yeardelta = year - 533
        leapdays = trunc(yeardelta / 4)
        return EPOCH_SARKAWAG + (365 * yeardelta) + leapdays + yeardays

    return EPOCH + (365 * year) + yeardays


def from_jd(jd, method=None):
    """Convert a Julian day count to an Armenian date. Use the method of Sarkawag if requested."""
    if method == "sarkawag":
        dc = jd - EPOCH_SARKAWAG
        if dc < 0:
            raise ValueError("Day count out of range for method")
        years = trunc(dc / 365.25)
        yeardays = dc - (365 * years + trunc(years / 4))
        if yeardays == 0:
            yeardays = 366 if years % 4 == 0 else 365
            years -= 1
        months = trunc((yeardays - 1) / 30)
        days = yeardays - (30 * months)
        return years + 533, months + 1, trunc(days)

    dc = jd - EPOCH

    if dc < 0:
        raise ValueError("Day count out of range")

    years = trunc((dc - 1) / 365)
    months = trunc(((dc - 1) % 365) / 30)
    days = dc - (365 * years) - (30 * months)

    return years, months + 1, trunc(days)


def to_julian(year, month, day, method=None):
    return julian.from_jd(to_jd(year, month, day, method))


def from_julian(year, month, day, method=None):
    return from_jd(julian.to_jd(year, month, day), method)


def to_gregorian(year, month, day, method=None):
    return gregorian.from_jd(to_jd(year, month, day, method))


def from_gregorian(year, month, day, method=None):
    return from_jd(gregorian.to_jd(year, month, day), method)


def month_length(year, month, method=None):
    if month > 13:
        raise ValueError("Requested month %d doesn't exist" % month)

    if month == 13:
        return 6 if (method == "sarkawag" and leap(year)) else 5

    return 30


def monthcalendar(year, month, method=None):
    start_weekday = jwday(to_jd(year, month, 1, method))
    monthlen = month_length(year, month, method)
    return monthcalendarhelper(start_weekday, monthlen)


def format(year, month, day, lang=None):
    """Convert an Armenian date into a string with the format DD MONTH YYYY."""
    # pylint: disable=redefined-builtin
    lang = lang or "en"
    if lang[0:2] == 'hy' or lang[0:2] == 'am' or lang == 'arm':
        month_name = MONTHS_ARM[month - 1]
    else:
        month = month_name = MONTHS[month - 1]

    return "{0:d} {1:} {2:d}".format(day, month_name, year)


def tostring(year, month, day, lang=None):
    """Kept for backwards compatibility, the format function name will be standard across the library"""
    return format(year, month, day, lang)



# -*- coding: utf-8 -*-
# This file is part of convertdate.
# http://github.com/fitnr/convertdate
# Licensed under the MIT license:
# http://opensource.org/licenses/MIT
# Copyright (c) 2016, fitnr <fitnr@fakeisthenewreal>
from calendar import isleap
from datetime import date
from math import trunc

from . import gregorian, ordinal
from .utils import jwday, n_weeks

MON = 0
TUE = 1
WED = 2
THU = 3
FRI = 4
SAT = 5
SUN = 6


def to_jd(year, week, day):
    '''Return Julian day count of given ISO year, week, and day'''
    return day + n_weeks(SUN, gregorian.to_jd(year - 1, 12, 28), week)


def from_jd(jd):
    '''Return tuple of ISO (year, week, day) for Julian day'''
    return from_gregorian(*gregorian.from_jd(jd))


def weeks_per_year(year):
    '''Number of ISO weeks in a year'''
    # 53 weeks: any year starting on Thursday and any leap year starting on Wednesday
    jan1 = jwday(gregorian.to_jd(year, 1, 1))

    if jan1 == THU or (jan1 == WED and isleap(year)):
        return 53

    return 52


def from_gregorian(year, month, day):
    return date(year, month, day).isocalendar()


def to_gregorian(year, week, day):
    return gregorian.from_jd(to_jd(year, week, day))


def format(year, week, day):
    # pylint: disable=redefined-builtin
    return "{}-W{:02}-{}".format(year, week, day)



# -*- coding: utf-8 -*-
# This file is part of convertdate.
# http://github.com/fitnr/convertdate
# Licensed under the MIT license:
# http://opensource.org/licenses/MIT
# Copyright (c) 2016, fitnr <fitnr@fakeisthenewreal>
from . import gregorian, julian, julianday


class DayCount:

    '''A day count converter for the given epoch (in terms of Julian Day Count)'''

    def __init__(self, epoch):
        self.epoch = epoch

    def to_gregorian(self, dc):
        '''From day count to Gregorian (year, month, day) tuple'''
        return gregorian.from_jd(self.to_jd(dc))

    def from_gregorian(self, year, month, day):
        return self.from_jd(gregorian.to_jd(year, month, day))

    def to_jd(self, dc):
        return dc + self.epoch

    def from_jd(self, jdc):
        return jdc - self.epoch

    def from_julian(self, year, month, day):
        return self.from_jd(julian.to_jd(year, month, day))

    def to_julian(self, dc):
        '''From day count to Julian (year, month, day) tuple'''
        return julian.from_jd(self.to_jd(dc))

    def to_datetime(self, dc):
        return julianday.to_datetime(self.to_jd(dc))

    def from_datetime(self, date):
        return self.from_jd(julianday.from_datetime(date))



# -*- coding: utf-8 -*-
# This file is part of convertdate.
# http:#github.com/fitnr/convertdate
# Licensed under the MIT license:
# http:#opensource.org/licenses/MIT
# Copyright (c) 2017, fitnr <fitnr@fakeisthenewreal>
"""
The `Coptic calendar <https://en.wikipedia.org/wiki/Coptic_calendar>`__,
also called the Alexandrian calendar, is a liturgical calendar used by the
Coptic Orthodox Church and some communities in Egypt. It is a reformed version
of the ancient Egyptian calendar.

It consists of twelve months of 30 days, followed by a "little month" of five days
(six in leap years).
"""
from . import gregorian
from .utils import floor, jwday, monthcalendarhelper

EPOCH = 1825029.5
MONTHS = [
    "Thout",
    "Paopi",
    "Hathor",
    "Koiak",
    "Tobi",
    "Meshir",
    "Paremhat",
    "Paremoude",
    "Pashons",
    "Paoni",
    "Epip",
    "Mesori",
    "Pi Kogi Enavot",
]
WEEKDAYS = ["Tkyriaka", "Pesnau", "Pshoment", "Peftoou", "Ptiou", "Psoou", "Psabbaton"]


def is_leap(year):
    "Determine whether this is a leap year."
    return year % 4 == 3 or year % 4 == -1


def to_jd(year, month, day):
    "Retrieve the Julian date equivalent for this date"
    return day + (month - 1) * 30 + (year - 1) * 365 + floor(year / 4) + EPOCH - 1


def from_jd(jdc):
    "Create a new date from a Julian date."
    cdc = floor(jdc) + 0.5 - EPOCH
    year = floor((cdc - floor((cdc + 366) / 1461)) / 365) + 1

    yday = jdc - to_jd(year, 1, 1)

    month = floor(yday / 30) + 1
    day = yday - (month - 1) * 30 + 1
    return int(year), int(month), int(day)


def to_gregorian(year, month, day):
    return gregorian.from_jd(to_jd(year, month, day))


def from_gregorian(year, month, day):
    return from_jd(gregorian.to_jd(year, month, day))


def month_length(year, month):
    if month <= 12:
        return 30

    if month == 13:
        if is_leap(year):
            return 6

        return 5

    raise ValueError("Invalid month")


def monthcalendar(year, month):
    start_weekday = jwday(to_jd(year, month, 1))
    monthlen = month_length(year, month)
    return monthcalendarhelper(start_weekday, monthlen)


def format(year, month, day):
    """Convert a Coptic date into a string with the format DD MONTH YYYY."""
    # pylint: disable=redefined-builtin
    return "{0:d} {1:} {2:d}".format(day, MONTHS[month - 1], year)



# -*- coding: utf-8 -*-
# This file is part of convertdate.
# http://github.com/fitnr/convertdate
# http://opensource.org/licenses/MIT
# Copyright (c) 2016, fitnr <fitnr@fakeisthenewreal>
# Licensed under the MIT license
"""
The `French Republican calendar <https://en.wikipedia.org/wiki/French_Republican_calendar>`__
was created during the heroic overthrow of the Ancien Regime.

Leap year calculations in the French Republican calendar are a matter of
dispute. By default, `convertdate` calculates leap years using the
autumnal equinox. You can also use one of three more systematic methods
proposed over the years.

-   Romme, a co-creator of the calendar, proposed leap years in years
    divisible by four, except for years divisible by 100.
-   Some concordances were drawn up in the 19th century that gave leap
    years every 4 years, in years that give a remainder of three when
    divided by four (19, 23, 27, etc...).
-   Von Mädler proposed leap years in years divisible by four, except
    for years divisible by 128.

You can specify any of these three methods with the method keyword
argument in `french_republican` conversion functions.

.. code-block:: python

   from convertdate import french_republican

   french_republican.to_gregorian(20, 1, 1), method='romme')
   # (1811, 9, 23)

   french_republican.to_gregorian(20, 1, 1), method='continuous')
   # (1811, 9, 24)

   french_republican.to_gregorian(20, 1, 1), method='madler')
   # (1811, 9, 23)

All the conversion methods correctly assign the leap years implemented
while calendar was in use (3, 7, 11).
"""
from math import trunc

from pymeeus.Sun import Sun

from . import gregorian
from .data.french_republican_days import french_republican_days

# julian day (1792, 9, 22)
EPOCH = 2375839.5

YEAR_EPOCH = 1791.0

DAYS_IN_YEAR = 365.0

MOIS = MONTHS = [
    "Vendémiaire",
    "Brumaire",
    "Frimaire",
    "Nivôse",
    "Pluviôse",
    "Ventôse",
    "Germinal",
    "Floréal",
    "Prairial",
    "Messidor",
    "Thermidor",
    "Fructidor",
    "Sansculottides",
]

LEAP_CYCLE_DAYS = 1461.0  # 365 * 4 + 1
LEAP_CYCLE_YEARS = 4.0


def leap(year, method=None):
    """
    Determine if this is a leap year in the FR calendar using one of three methods: 4, 100, 128
    (every 4th years, every 4th or 400th but not 100th, every 4th but not 128th)

    Methods:
        * 4 (concordance rule): leap every four years: 3, 7, 11, 15, ... etc
        * 100 (Romme's rule): leap every 4th and 400th year, but not 100th:
            20, 24, ... 96, 104, ... 396, 400, 404 ...
        * 128 (von Mädler's rule): leap every 4th but not 128th: 20, 24, ... 124, 132, ...
        * equinox [default]: use calculation of the equinox to determine date, never returns a leap year
    """
    method = method or 'equinox'

    if year in (3, 7, 11):
        return True

    if year < 15:
        return False

    if method in (4, 'continuous') or (year <= 16 and method in (128, 'madler', 4, 'continuous')):
        return year % 4 == 3

    if method in (100, 'romme'):
        return (year % 4 == 0 and year % 100 != 0) or year % 400 == 0

    if method in (128, 'madler'):
        return year % 4 == 0 and year % 128 != 0

    if method == 'equinox':
        # Is equinox on 366th day after (year, 1, 1)
        startjd = to_jd(year, 1, 1, method='equinox')
        if premier_da_la_annee(startjd + 367) - startjd == 366.0:
            return True
    else:
        raise ValueError("Unknown leap year method. Try: continuous, romme, madler or equinox")

    return False


def _previous_fall_equinox(jd):
    '''Return the julian day count of the previous fall equinox.'''
    y, _, _ = gregorian.from_jd(jd)
    eqx = Sun.get_equinox_solstice(y, "autumn").jde()
    if eqx > jd:
        eqx = Sun.get_equinox_solstice(y - 1, "autumn").jde()

    return eqx


def _next_fall_equinox(jd):
    '''Return the julian day count of the previous fall equinox.'''
    y, _, _ = gregorian.from_jd(jd)
    eqx = Sun.get_equinox_solstice(y, "autumn").jde()
    if eqx < jd:
        eqx = Sun.get_equinox_solstice(y + 1, "autumn").jde()

    return eqx


def premier_da_la_annee(jd):
    """
    Returns Julian day number containing fall equinox (first day of the FR year)
    of the current FR year.
    """
    previous = trunc(_previous_fall_equinox(jd) - 0.5) + 0.5

    if previous + 364 < jd:
        # test if current day is the equinox if the previous equinox was a long time ago
        nxt = trunc(_next_fall_equinox(jd) - 0.5) + 0.5

        if nxt <= jd:
            return nxt

    return previous


def to_jd(year, month, day, method=None):
    '''Obtain Julian day from a given French Revolutionary calendar date.'''
    method = method or 'equinox'

    if day < 1 or day > 30:
        raise ValueError("Invalid day for this calendar")

    if month > 13:
        raise ValueError("Invalid month for this calendar")

    if month == 13 and day > 5 + leap(year, method=method):
        raise ValueError("Invalid day for this month in this calendar")

    if method == 'equinox':
        return _to_jd_equinox(year, month, day)

    return _to_jd_schematic(year, month, day, method)


def _to_jd_schematic(year, month, day, method):
    '''Calculate JD using various leap-year calculation methods'''
    y0, y1, y2, y3, y4, y5 = 0, 0, 0, 0, 0, 0

    intercal_cycle_yrs, over_cycle_yrs, leap_suppression_yrs = None, None, None

    # Use the every-four-years method below year 16 (madler) or below 15 (romme)
    if (method in (100, 'romme') and year < 15) or (method in (128, 'madler') and year < 17):
        method = 4

    if method in (4, 'continuous'):
        # Leap years: 15, 19, 23, ...
        y5 = -365

    elif method in (100, 'romme'):
        year = year - 13
        y5 = DAYS_IN_YEAR * 12 + 3

        leap_suppression_yrs = 100.0
        leap_suppression_days = 36524  # leap_cycle_days * 25 - 1

        intercal_cycle_yrs = 400.0
        intercal_cycle_days = 146097  # leap_suppression_days * 4 + 1

        over_cycle_yrs = 4000.0
        over_cycle_days = 1460969  # intercal_cycle_days * 10 - 1

    elif method in (128, 'madler'):
        year = year - 17
        y5 = DAYS_IN_YEAR * 16 + 4

        leap_suppression_days = 46751  # 32 * leap_cycle_days - 1
        leap_suppression_yrs = 128

    else:
        raise ValueError("Unknown leap year method. Try: continuous, romme, madler or equinox")

    if over_cycle_yrs:
        y0 = trunc(year / over_cycle_yrs) * over_cycle_days
        year = year % over_cycle_yrs

    # count intercalary cycles in days (400 years long or None)
    if intercal_cycle_yrs:
        y1 = trunc(year / intercal_cycle_yrs) * intercal_cycle_days
        year = year % intercal_cycle_yrs

    # count leap suppresion cycles in days (100 or 128 years long)
    if leap_suppression_yrs:
        y2 = trunc(year / leap_suppression_yrs) * leap_suppression_days
        year = year % leap_suppression_yrs

    y3 = trunc(year / LEAP_CYCLE_YEARS) * LEAP_CYCLE_DAYS
    year = year % LEAP_CYCLE_YEARS

    # Adjust 'year' by one to account for lack of year 0
    y4 = year * DAYS_IN_YEAR

    yj = y0 + y1 + y2 + y3 + y4 + y5

    mj = (month - 1) * 30

    return EPOCH + yj + mj + day - 1


def _to_jd_equinox(an, mois, jour):
    '''Return jd of this FR date, counting from the previous equinox.'''
    day_of_adr = (30 * (mois - 1)) + (jour - 1)
    equinoxe = _next_fall_equinox(gregorian.to_jd(an + YEAR_EPOCH, 1, 1))
    return trunc(equinoxe - 0.5) + 0.5 + day_of_adr


def from_jd(jd, method=None):
    """Calculate date in the French Revolutionary
    calendar from Julian day.  The five or six
    "sansculottides" are considered a thirteenth
    month in the results of this function."""
    method = method or 'equinox'

    if method == 'equinox':
        return _from_jd_equinox(jd)

    return _from_jd_schematic(jd, method)


def _from_jd_schematic(jd, method):
    '''Convert from JD using various leap-year calculation methods'''
    if jd < EPOCH:
        raise ValueError("Can't convert days before the French Revolution")

    # days since Epoch
    J = trunc(jd) + 0.5 - EPOCH

    y0, y1, y2, y3, y4, y5 = 0, 0, 0, 0, 0, 0
    intercal_cycle_days = leap_suppression_days = over_cycle_days = None

    # Use the every-four-years method below year 17
    if (J <= DAYS_IN_YEAR * 12 + 3 and method in (100, 'romme')) or (
        J <= DAYS_IN_YEAR * 17 + 4 and method in (128, 'madler')
    ):
        method = 4

    # set p and r in Hatcher algorithm
    if method in (4, 'continuous'):
        # Leap years: 15, 19, 23, ...
        # Reorganize so that leap day is last day of cycle
        J = J + 365
        y5 = -1

    elif method in (100, 'romme'):
        # Year 15 is not a leap year
        # Year 16 is leap, then multiples of 4, not multiples of 100, yes multiples of 400
        y5 = 12
        J = J - DAYS_IN_YEAR * 12 - 3

        leap_suppression_yrs = 100.0
        leap_suppression_days = 36524  # LEAP_CYCLE_DAYS * 25 - 1

        intercal_cycle_yrs = 400.0
        intercal_cycle_days = 146097  # leap_suppression_days * 4 + 1

        over_cycle_yrs = 4000.0
        over_cycle_days = 1460969  # intercal_cycle_days * 10 - 1

        y0 = trunc(J / over_cycle_days) * over_cycle_yrs
        J = J % over_cycle_days

        y1 = trunc(J / intercal_cycle_days) * intercal_cycle_yrs
        J = J % intercal_cycle_days

    elif method in (128, 'madler'):
        # Year 15 is a leap year, then year 20 and multiples of 4, not multiples of 128
        y5 = 16
        J = J - DAYS_IN_YEAR * 16 - 4

        leap_suppression_yrs = 128
        leap_suppression_days = 46751  # 32 * leap_cycle_days - 1

    else:
        raise ValueError("Unknown leap year method. Try: continuous, romme, madler or equinox")

    if leap_suppression_days:
        y2 = trunc(J / leap_suppression_days) * leap_suppression_yrs
        J = J % leap_suppression_days

    y3 = trunc(J / LEAP_CYCLE_DAYS) * LEAP_CYCLE_YEARS

    if J % LEAP_CYCLE_DAYS == LEAP_CYCLE_DAYS - 1:
        J = 1460
    else:
        J = J % LEAP_CYCLE_DAYS

    # 0 <= J <= 1460
    # J needs to be 365 here on leap days ONLY

    y4 = trunc(J / DAYS_IN_YEAR)

    if J == DAYS_IN_YEAR * 4:
        y4 = y4 - 1
        J = 365.0
    else:
        J = J % DAYS_IN_YEAR

    year = y0 + y1 + y2 + y3 + y4 + y5

    month = trunc(J / 30.0)
    J = J - month * 30

    return trunc(year) + 1, month + 1, trunc(J) + 1


def _from_jd_equinox(jd):
    '''Calculate the FR day using the equinox as day 1'''
    jd = trunc(jd) + 0.5
    equinoxe = premier_da_la_annee(jd)

    an = int(gregorian.from_jd(equinoxe)[0] - YEAR_EPOCH)
    mois = trunc((jd - equinoxe) / 30.0) + 1
    jour = int((jd - equinoxe) % 30) + 1

    return (an, mois, jour)


def decade(jour):
    return trunc(jour / 100.0) + 1


def day_name(month, day):
    return french_republican_days[month][day - 1]


def from_gregorian(year, month, day, method=None):
    return from_jd(gregorian.to_jd(year, month, day), method=method)


def to_gregorian(an, mois, jour, method=None):
    return gregorian.from_jd(to_jd(an, mois, jour, method=method))


def format(an, mois, jour):
    """Convert a FR date into a string with the format DD MONTH YYYY."""
    # pylint: disable=redefined-builtin
    return "{0} {1} {2}".format(jour, MOIS[mois - 1], an)



# -*- coding: utf-8 -*-
# This file is part of convertdate.
# http://github.com/fitnr/convertdate
# Licensed under the MIT license:
# http://opensource.org/licenses/MIT
# Copyright (c) 2016, 2020, 2021 fitnr <fitnr@fakeisthenewreal>
"""
The Convertdate library contains methods and functions for converting dates between
different calendar systems.

It was originally developed as as `Python Date Util <(http://sourceforge.net/projects/pythondateutil/>`__
by Phil Schwartz. It had been significantly updated and expanded.

Most of the original code is ported from
`Fourmilab's calendar converter <http://www.fourmilab.ch/documents/calendar/>`__,
which was developed by John Walker.

The algorithms are believed to be derived from: Meeus, Jean. `Astronomical Algorithms`,
Richmond: Willmann-Bell, 1991 (ISBN 0-943396-35-2)
"""
from . import armenian
from . import bahai
from . import coptic
from . import daycount
from . import dublin
from . import french_republican
from . import gregorian
from . import hebrew
from . import holidays
from . import indian_civil
from . import iso
from . import islamic
from . import julian
from . import julianday
from . import mayan
from . import persian
from . import positivist
from . import ordinal
from . import utils

__version__ = '2.3.2'

__all__ = [
    'holidays',
    'armenian',
    'bahai',
    'coptic',
    'dublin',
    'daycount',
    'french_republican',
    'gregorian',
    'hebrew',
    'indian_civil',
    'islamic',
    'iso',
    'julian',
    'julianday',
    'mayan',
    'persian',
    'positivist',
    'mayan',
    'ordinal',
    'utils',
]



# -*- coding: utf-8 -*-

# This file is part of convertdate.
# http://github.com/fitnr/convertdate

# Licensed under the MIT license:
# http://opensource.org/licenses/MIT
# Copyright (c) 2016, fitnr <fitnr@fakeisthenewreal>

# source: Wikipedia
# month 13 is the Sansculottides
french_republican_days = {
    1: [
        "Raisin",
        "Safran",
        "Châtaigne",
        "Colchique",
        "Cheval",
        "Balsamine",
        "Carotte",
        "Amaranthe",
        "Panais",
        "Cuve",
        "Pomme de terre",
        "Immortelle",
        "Potiron",
        "Réséda",
        "Âne",
        "Belle de nuit",
        "Citrouille",
        "Sarrasin",
        "Tournesol",
        "Pressoir",
        "Chanvre",
        "Pêche",
        "Navet",
        "Amaryllis",
        "Bœuf",
        "Aubergine",
        "Piment",
        "Tomate",
        "Orge",
        "Tonneau",
    ],
    2: [
        "Pomme",
        "Céleri",
        "Poire",
        "Betterave",
        "Oie",
        "Héliotrope",
        "Figue",
        "Scorsonère",
        "Alisier",
        "Charrue",
        "Salsifis",
        "Mâcre",
        "Topinambour",
        "Endive",
        "Dindon",
        "Chervis",
        "Cresson",
        "Dentelaire",
        "Grenade",
        "Herse",
        "Bacchante",
        "Azerole",
        "Garance",
        "Orange",
        "Faisan",
        "Pistache",
        "Macjonc",
        "Coing",
        "Cormier",
        "Rouleau",
    ],
    3: [
        "Raiponce",
        "Turneps",
        "Chicorée",
        "Nèfle",
        "Cochon",
        "Mâche",
        "Chou-fleur",
        "Miel",
        "Genièvre",
        "Pioche",
        "Cire",
        "Raifort",
        "Cèdre",
        "Sapin",
        "Chevreuil",
        "Ajonc",
        "Cyprès",
        "Lierre",
        "Sabine",
        "Hoyau",
        "Érable à sucre",
        "Bruyère",
        "Roseau",
        "Oseille",
        "Grillon",
        "Pignon",
        "Liège",
        "Truffe",
        "Olive",
        "Pelle",
    ],
    4: [
        "Tourbe",
        "Houille",
        "Bitume",
        "Soufre",
        "Chien",
        "Lave",
        "Terre végétale",
        "Fumier",
        "Salpêtre",
        "Fléau",
        "Granit",
        "Argile",
        "Ardoise",
        "Grès",
        "Lapin",
        "Silex",
        "Marne",
        "Pierre à chaux",
        "Marbre",
        "Van",
        "Pierre à plâtre",
        "Sel",
        "Fer",
        "Cuivre",
        "Chat",
        "Étain",
        "Plomb",
        "Zinc",
        "Mercure",
        "Crible",
    ],
    5: [
        "Lauréole",
        "Mousse",
        "Fragon",
        "Perce-neige",
        "Taureau",
        "Laurier-thym",
        "Amadouvier",
        "Mézéréon",
        "Peuplier",
        "Coignée",
        "Ellébore",
        "Brocoli",
        "Laurier",
        "Avelinier",
        "Vache",
        "Buis",
        "Lichen",
        "If",
        "Pulmonaire",
        "Serpette",
        "Thlaspi",
        "Thimelé",
        "Chiendent",
        "Trainasse",
        "Lièvre",
        "Guède",
        "Noisetier",
        "Cyclamen",
        "Chélidoine",
        "Traîneau",
    ],
    6: [
        "Tussilage",
        "Cornouiller",
        "Violier",
        "Troène",
        "Bouc",
        "Asaret",
        "Alaterne",
        "Violette",
        "Marceau",
        "Bêche",
        "Narcisse",
        "Orme",
        "Fumeterre",
        "Vélar",
        "Chèvre",
        "Épinard",
        "Doronic",
        "Mouron",
        "Cerfeuil",
        "Cordeau",
        "Mandragore",
        "Persil",
        "Cochléaria",
        "Pâquerette",
        "Thon",
        "Pissenlit",
        "Sylvie",
        "Capillaire",
        "Frêne",
        "Plantoir",
    ],
    7: [
        "Primevère",
        "Platane",
        "Asperge",
        "Tulipe",
        "Poule",
        "Bette",
        "Bouleau",
        "Jonquille",
        "Aulne",
        "Couvoir",
        "Pervenche",
        "Charme",
        "Morille",
        "Hêtre",
        "Abeille",
        "Laitue",
        "Mélèze",
        "Ciguë",
        "Radis",
        "Ruche",
        "Gainier",
        "Romaine",
        "Marronnier",
        "Roquette",
        "Pigeon",
        "Lilas",
        "Anémone",
        "Pensée",
        "Myrtille",
        "Greffoir",
    ],
    8: [
        "Rose",
        "Chêne",
        "Fougère",
        "Aubépine",
        "Rossignol",
        "Ancolie",
        "Muguet",
        "Champignon",
        "Hyacinthe",
        "Râteau",
        "Rhubarbe",
        "Sainfoin",
        "Bâton d'or",
        "Chamerisier",
        "Ver à soie",
        "Consoude",
        "Pimprenelle",
        "Corbeille d'or",
        "Arroche",
        "Sarcloir",
        "Statice",
        "Fritillaire",
        "Bourrache",
        "Valériane",
        "Carpe",
        "Fusain",
        "Civette",
        "Buglosse",
        "Sénevé",
        "Houlette",
    ],
    9: [
        "Luzerne",
        "Hémérocalle",
        "Trèfle",
        "Angélique",
        "Canard",
        "Mélisse",
        "Fromental",
        "Martagon",
        "Serpolet",
        "Faux",
        "Fraise",
        "Bétoine",
        "Pois",
        "Acacia",
        "Caille",
        "Œillet",
        "Sureau",
        "Pavot",
        "Tilleul",
        "Fourche",
        "Barbeau",
        "Camomille",
        "Chèvrefeuille",
        "Caille-lait",
        "Tanche",
        "Jasmin",
        "Verveine",
        "Thym",
        "Pivoine",
        "Chariot",
    ],
    10: [
        "Seigle",
        "Avoine",
        "Oignon",
        "Véronique",
        "Mulet",
        "Romarin",
        "Concombre",
        "Échalote",
        "Absinthe",
        "Faucille",
        "Coriandre",
        "Artichaut",
        "Girofle",
        "Lavande",
        "Chamois",
        "Tabac",
        "Groseille",
        "Gesse",
        "Cerise",
        "Parc",
        "Menthe",
        "Cumin",
        "Haricot",
        "Orcanète",
        "Pintade",
        "Sauge",
        "Ail",
        "Vesce",
        "Blé",
        "Chalémie",
    ],
    11: [
        "Épeautre",
        "Bouillon blanc",
        "Melon",
        "Ivraie",
        "Bélier",
        "Prêle",
        "Armoise",
        "Carthame",
        "Mûre",
        "Arrosoir",
        "Panic",
        "Salicorne",
        "Abricot",
        "Basilic",
        "Brebis",
        "Guimauve",
        "Lin",
        "Amande",
        "Gentiane",
        "Écluse",
        "Carline",
        "Câprier",
        "Lentille",
        "Aunée",
        "Loutre",
        "Myrte",
        "Colza",
        "Lupin",
        "Coton",
        "Moulin",
    ],
    12: [
        "Prune",
        "Millet",
        "Lycoperdon",
        "Escourgeon",
        "Saumon",
        "Tubéreuse",
        "Sucrion",
        "Apocyn",
        "Réglisse",
        "Échelle",
        "Pastèque",
        "Fenouil",
        "Épine vinette",
        "Noix",
        "Truite",
        "Citron",
        "Cardère",
        "Nerprun",
        "Tagette",
        "Hotte",
        "Églantier",
        "Noisette",
        "Houblon",
        "Sorgho",
        "Écrevisse",
        "Bigarade",
        "Verge d'or",
        "Maïs",
        "Marron",
        "Panier",
    ],
    13: [
        "La Fête de la Vertu",
        "La Fête du Génie",
        "La Fête du Travail",
        "La Fête de l'Opinion",
        "La Fête des Récompenses",
        "La Fête de la Révolution",
    ],
}



# -*- coding: utf-8 -*-
# This file is part of convertdate.
# http://github.com/fitnr/convertdate
# Licensed under the MIT license:
# http://opensource.org/licenses/MIT
# Copyright (c) 2016, fitnr <fitnr@fakeisthenewreal>

"""
The Bahá'í (Badí) calendar is a solar calendar with 19 months of 19 days.

Every four years, an intercalary period, Ayyam-i-Há, occurs between the 18th and 19th
months. Dates in this period are returned as month 19, and the month of ‘Alá is always
reported as month 20.

.. code-block:: python

   from convertdate import bahai
   # the first day of Ayyam-i-Ha:
   bahai.to_gregorian(175, 19, 1)
   # (2019, 2, 26)
   # The first day of 'Ala:
   bahai.to_gregorian(175, 20, 1)
   # (2019, 3, 2)

"""
from calendar import isleap
from math import ceil, trunc

from pymeeus.Angle import Angle
from pymeeus.Epoch import Epoch
from pymeeus.Sun import Sun

from . import gregorian
from .utils import jwday, monthcalendarhelper

EPOCH = 2394646.5
EPOCH_GREGORIAN_YEAR = 1844

TEHRAN = 51.4215, 35.6944

WEEKDAYS = ("Jamál", "Kamál", "Fidál", "Idál", "Istijlál", "Istiqlál", "Jalál")

MONTHS = (
    "Bahá",
    "Jalál",
    "Jamál",
    "‘Aẓamat",
    "Núr",
    "Raḥmat",
    "Kalimát",
    "Kamál",
    "Asmá’",
    "‘Izzat",
    "Mashíyyat",
    "‘Ilm",
    "Qudrat",
    "Qawl",
    "Masá’il",
    "Sharaf",
    "Sulṭán",
    "Mulk",
    "Ayyám-i-Há",
    "‘Alá",
)

ENGLISH_MONTHS = (
    "Splendor",
    "Glory",
    "Beauty",
    "Grandeur",
    "Light",
    "Mercy",
    "Words",
    "Perfection",
    "Names",
    "Might",
    "Will",
    "Knowledge",
    "Power",
    "Speech",
    "Questions",
    "Honour",
    "Sovereignty",
    "Dominion",
    "Days of Há",
    "Loftiness",
)

BAHA = 1
JALAL = 2
JAMAL = 3
AZAMAT = 4
NUR = 5
RAHMAT = 6
KALIMAT = 7
KAMAL = 8
ASMA = 9
IZZAT = 10
MASHIYYAT = 11
ILM = 12
QUDRAT = 13
QAWL = 14
MASAIL = 15
SHARAF = 16
SULTAN = 17
MULK = 18
AYYAMIHA = 19
ALA = 20


def gregorian_nawruz(year):
    """
    Return Nawruz in the Gregorian calendar.
    Returns a tuple (month, day), where month is always 3
    """
    if year == 2059:
        return 3, 20

    # Timestamp of spring equinox.
    equinox = Sun.get_equinox_solstice(year, "spring")

    # Get times of sunsets in Tehran near vernal equinox.
    x, y = Angle(TEHRAN[0]), Angle(TEHRAN[1])
    days = trunc(equinox.get_date()[2]), ceil(equinox.get_date()[2])

    for day in days:
        sunset = Epoch(year, 3, day).rise_set(y, x)[1]
        if sunset > equinox:
            return 3, day

    raise ValueError("Couldn't find date of Nawruz.")


def to_jd(year, month, day):
    '''Determine Julian day from Bahai date'''
    if month <= 18:
        gy = year - 1 + EPOCH_GREGORIAN_YEAR
        n_month, n_day = gregorian_nawruz(gy)
        return gregorian.to_jd(gy, n_month, n_day - 1) + day + (month - 1) * 19
    if month == 19:
        # Count Ayyám-i-Há from the last day of Mulk
        return to_jd(year, month - 1, 19) + day
    # For the month of ‘Alá we will count _backwards_ from the next Naw Rúz
    gy = year + EPOCH_GREGORIAN_YEAR
    n_month, n_day = gregorian_nawruz(gy)
    return gregorian.to_jd(gy, n_month, n_day) - 20 + day


def from_jd(jd):
    '''Calculate Bahai date from Julian day'''
    jd = trunc(jd) + 0.5
    g = gregorian.from_jd(jd)
    gy = g[0]
    n_month, n_day = gregorian_nawruz(gy)

    bstarty = EPOCH_GREGORIAN_YEAR

    if jd <= gregorian.to_jd(gy, n_month, 20):
        x = 1
    else:
        x = 0
    # verify this next line...
    bys = gy - (bstarty + (((gregorian.to_jd(gy, 1, 1) <= jd) and x)))

    year = bys + 1
    days = jd - to_jd(year, 1, 1)
    bld = to_jd(year, n_day - 1, 1)

    if jd >= bld:
        month = 20
    else:
        month = trunc(days / 19) + 1
    day = int((jd + 1) - to_jd(year, month, 1))

    return year, month, day


def from_gregorian(year, month, day):
    return from_jd(gregorian.to_jd(year, month, day))


def to_gregorian(year, month, day):
    return gregorian.from_jd(to_jd(year, month, day))


def month_length(year, month):
    gy = year + EPOCH_GREGORIAN_YEAR - 1

    if month == 19:
        _, nawruz_future = gregorian_nawruz(gy + 1)
        _, nawruz_past = gregorian_nawruz(gy)
        length_of_year = nawruz_future + 365 - nawruz_past

        if isleap(gy + 1):
            length_of_year = length_of_year + 1

        return length_of_year - 19 * 19

    return 19


def monthcalendar(year, month):
    start_weekday = jwday(to_jd(year, month, 1))
    monthlen = month_length(year, month)
    return monthcalendarhelper(start_weekday, monthlen)


def format(year, month, day, lang=None):
    """Convert a Baha'i date into a string with the format DD MONTH YYYY."""
    # pylint: disable=redefined-builtin
    lang = lang or "en"
    if lang[0:2] == 'ar' or lang[0:2] == 'fa':
        month_name = MONTHS[month - 1]
    else:
        month_name = ENGLISH_MONTHS[month - 1]

    return "{0:d} {1:} {2:d}".format(day, month_name, year)



# -*- coding: utf-8 -*-
# This file is part of convertdate.
# http://github.com/fitnr/convertdate
# Licensed under the MIT license:
# http://opensource.org/licenses/MIT
# Copyright (c) 2016, fitnr <fitnr@fakeisthenewreal>
"""
The `ordinal date <https://en.wikipedia.org/wiki/Ordinal_date>` specifies the day
of year as a number between 1 and 366.

Ordinal dates are represented by a tuple: ``(year, dayofyear)``
"""
from calendar import isleap
from math import trunc

from . import gregorian


def to_jd(year, dayofyear):
    '''Return Julian day count of given ordinal date.'''
    return gregorian.to_jd(year, 1, 1) + dayofyear - 1


def from_jd(jd):
    '''Convert a Julian day count to an ordinal date.'''
    year, _, _ = gregorian.from_jd(jd)
    return year, round(jd - gregorian.to_jd(year, 1, 1) + 1)


def from_gregorian(year, month, day):
    """Convert a Gregorian date to an ordinal date."""
    m = month + 1

    if m <= 3:
        m = m + 12

    leap = isleap(year)

    t = trunc(30.6 * m) + day - 122 + 59 + leap

    if t > 365 + leap:
        t = t - 365 - leap

    return year, t


def to_gregorian(year, dayofyear):
    """Convert an ordinal date to a Gregorian date."""
    leap = isleap(year)

    if dayofyear < 59 + leap:
        leap_adj = 0
    elif leap:
        leap_adj = 1
    else:
        leap_adj = 2

    month = trunc((((dayofyear - 1 + leap_adj) * 12) + 373) / 367)

    startofmonth = from_gregorian(year, month, 1)

    return year, month, dayofyear - startofmonth[1] + 1
